"""
module mqkit.consume.decorator

Defines the consume decorator for designating functions as message queue consumers.
This is a single-consumer alternative to the mqkit.apps.App class.
"""

import inspect
from logging import Logger
import logging
import os
import sys
from typing import Callable, NoReturn, Optional, Union

from mqkit.endpoints.endpoint import EndpointExceptionHandler
from mqkit.messaging.forwardtarget import ForwardTarget

from ..engines import create_engine, Engine
from ..endpoints import EndpointFactory
from ..endpoints.config import QueueEndpointConfig
from ..marshal.codecs import CodecType
from ..messaging import Destination, Exchange, Queue
from ..messaging.retry import NoRetryStrategy, RetryStrategy
from ..workers.threaded import ThreadWorker


def _consume_threaded(
    config: QueueEndpointConfig,
    engine: Engine,
    logger: Logger,
) -> None:
    # instantiate the endpoint and start a threaded worker
    worker: ThreadWorker = ThreadWorker(
        EndpointFactory.create_queue_endpoint(config),
        engine=engine,
        error_queue=None,
    )
    try:
        worker.start()
        logger.info("Started threaded consumer for queue '%s'", config.queue.name)
        worker.join()
    except KeyboardInterrupt:  # pragma: no cover
        worker.stop("Keyboard interrupt received")
        worker.join()

    # if an exception occurred in the worker, re-raise it here
    if worker.error is not None:
        logger.exception(
            "Worker for queue '%s' exited with exception: %s",
            config.queue.name,
            worker.error,
            exc_info=worker.error,
        )
        raise worker.error


def _infer_engine() -> Engine:
    # try to infer the engine from an environment variable
    if "MQKIT_ENGINE_URL" not in os.environ:
        raise RuntimeError(
            "No engine provided and MQKIT_ENGINE_URL environment variable not set"
        )

    return create_engine(os.environ["MQKIT_ENGINE_URL"])


def _infer_logger(func: Callable) -> Logger:
    return logging.getLogger(
        os.environ.get(
            "MQKIT_CONSUMER_LOGGER_NAME",
            func.__name__,
        )
    )


def consume(
    # pylint: disable=too-many-arguments,duplicate-code
    name: str,
    *,
    engine: Optional[Engine] = None,
    logger: Optional[Logger] = None,
    codec: Optional[CodecType | str] = None,
    forward_to: Optional[Union[str, Queue, Exchange, Destination]] = None,
    persistent: bool = True,
    auto_delete: bool = False,
    retry_strategy: Optional[RetryStrategy] = None,
    dead_letter: Optional[ForwardTarget] = None,
    on_decode_error: Optional[EndpointExceptionHandler] = None,
    on_validation_error: Optional[EndpointExceptionHandler] = None,
) -> Callable[[Callable], NoReturn]:
    """
    Blocking decorator to designate a callback function as a consumer for a single message queue.
    Intended for use in applications that do not require single-process concurrency (for example,
    within a scaling container like a Kubernetes pod).

    Args:
        name (str): The name of the queue.
        codec (Optional[CodecType | str]): The codec type for message serialization.
            Defaults to CodecType.JSON.
        forward_to (Optional[str]): The name of the queue to forward results to.
            Defaults to None.
        persistent (bool): Whether the queue is persistent. Defaults to True.
        auto_delete (bool): Whether the queue is auto-delete. Defaults to False.

    Returns:
        Callable[[Callable], QueueEndpoint]: The decorator function.

    Raises:
        TypeError: If the function is not compatible with the selected concurrency mode.
    """

    if engine is None:
        engine = _infer_engine()

    codec = CodecType(codec) if codec is not None else CodecType.JSON

    # if the user didn't provide an explicit retry strategy, use the default
    # strategy that never performs retries
    if retry_strategy is None:
        retry_strategy = NoRetryStrategy()

    def _consume_decorator(func: Callable) -> NoReturn:
        # if a logger was not provided, infer one from the environment
        nonlocal logger
        if logger is None:
            logger = _infer_logger(func)

        # NOTE: eventually add async functionality here
        if inspect.iscoroutinefunction(func):
            raise NotImplementedError(
                "Async functions are not supported by the @consume decorator"
            )

        # run in threaded mode. exit with code 0 if no exception raised (should only
        # exit on KeyboardInterrupt)
        _consume_threaded(
            config=QueueEndpointConfig(
                queue=Queue(
                    name=name,
                    persistent=persistent,
                    auto_delete=auto_delete,
                ),
                target=func,
                codec_type=codec,
                forward_to=forward_to,
                retry_strategy=retry_strategy,
                dead_letter=dead_letter,
                error_handlers=QueueEndpointConfig.make_error_handlers_dict(
                    on_decode_error=on_decode_error,
                    on_validation_error=on_validation_error,
                ),
            ),
            engine=engine,
            logger=logger,
        )
        sys.exit(0)

    return _consume_decorator



"""
module mqkit.credentials.plaincredentials

Defines the PlainCredentials class for storing plain text username and password.
"""

from .credentials import Credentials


class PlainCredentials(Credentials):
    """
    class PlainCredentials

    Stores plain text username and password for authentication.
    """

    username: str
    password: str



"""
module mqkit.errors.functiontypeerror

Module defining the FunctionTypeError exception for MQKit.
"""

from .apperror import AppError


class FunctionTypeError(AppError):
    """Exception indicating that a callback function has an incorrect type."""



"""
module mqkit.marshal.codecs.yamlcodec

Defines the YamlCodec class for encoding and decoding YAML data in message queues.
"""

import yaml

from .codec import Codec
from ...errors import DecodeError, EncodeError


class YamlCodec(Codec):
    """
    class YamlCodec

    A codec for encoding and decoding YAML data.
    """

    _encoding: str

    def __init__(self: "YamlCodec", encoding: str = "utf-8") -> None:
        self._encoding = encoding

    @property
    def content_type(self: "YamlCodec") -> str:
        return "application/yaml"

    def decode(self: "YamlCodec", data: bytes) -> dict:
        try:
            return yaml.safe_load(
                data.decode(self._encoding),
            )
        except yaml.YAMLError as exc:
            raise DecodeError("Failed to decode YAML data") from exc

    def encode(self: "YamlCodec", data: dict) -> bytes:
        try:
            return yaml.safe_dump(data).encode(self._encoding)
        except yaml.YAMLError as exc:
            raise EncodeError("Failed to encode YAML data") from exc



"""
module mqkit.messaging.exchangetype

Defines the ExchangeType enumeration for message queue exchanges.
"""

from enum import Enum


class ExchangeType(str, Enum):
    """
    class ExchangeType

    Enumeration of supported message queue exchange types.
    """

    DIRECT = "direct"
    FANOUT = "fanout"
    TOPIC = "topic"
    HEADERS = "headers"



"""
module mqkit.marshal.codecs.codectype

Defines the CodecType enumeration for supported serialization formats.
"""

from enum import Enum


class CodecType(str, Enum):
    """
    class CodecType

    Enumeration of supported serialization formats for message payloads.
    """

    JSON = "json"
    MESSAGEPACK = "messagepack"
    RAW = "raw"
    YAML = "yaml"



"""
module mqkit.marshal.codecs.jsoncodec

Defines the JsonCodec class for encoding and decoding JSON data.
"""

import json
from typing import Any, Dict

from .codec import Codec
from ...errors import DecodeError, EncodeError


class JsonCodec(Codec):
    """
    class JsonCodec

    A codec for encoding and decoding JSON data.
    """

    _encoding: str

    def __init__(self: "JsonCodec", encoding: str = "utf-8") -> None:
        self._encoding = encoding

    @property
    def content_type(self: "JsonCodec") -> str:
        return "application/json"

    def decode(self: "JsonCodec", data: bytes) -> Dict[str, Any]:
        try:
            return json.loads(
                data.decode(self._encoding),
            )
        except json.JSONDecodeError as exc:
            raise DecodeError("Failed to decode JSON data") from exc

    def encode(self: "JsonCodec", data: object) -> bytes:
        try:
            return json.dumps(data).encode(self._encoding)
        except (TypeError, ValueError) as exc:
            raise EncodeError("Failed to encode data to JSON") from exc



"""
module mqkit.marshal.forward

A module defining the Forward data model for message forwarding.
"""

from typing import Union

from pydantic import BaseModel

from .exchange import Exchange
from .queue import Queue
from .queuemessage import QueueMessage


class Forward(BaseModel):
    """
    class Forward

    Represents a forwarding instruction for a message, specifying the target
    queue and the message to be forwarded.
    """

    forward_target: Union[Queue, Exchange]
    message: QueueMessage



"""
module mqkit.credentials.credentials

Defines the Credentials abstract base class for representing credentials used
for authenticating with message queue services.
"""

from abc import ABCMeta

from pydantic import BaseModel


class Credentials(BaseModel, metaclass=ABCMeta):
    """
    class Credentials

    Abstract base class for representing credentials used for authenticating
    with message queue services.
    """

    def __repr__(self: "Credentials") -> str:
        return f"{self.__class__.__name__}(...)"

    def __str__(self: "Credentials") -> str:
        return self.__repr__()



"""
class ConfigurationWarning

A warning that is raised when a configuration issue is detected in the mqkit library. This
could include issues such as missing or invalid configuration parameters
"""

from .mqkitwarning import MqkitWarning


class ConfigurationWarning(MqkitWarning):
    """
    class ConfigurationWarning

    A warning that is raised when a configuration issue is detected in the mqkit library. This
    could include issues such as missing or invalid configuration parameters
    """



"""
module mqkit.messaging.exchange

Defines the Exchange class for message queue exchanges.
"""

from typing import Union
from pydantic import BaseModel

from .exchangetype import ExchangeType


class Exchange(BaseModel):
    """
    class Exchange

    Model representing a message queue exchange.
    """

    name: str
    type: ExchangeType
    persistent: bool = True
    auto_delete: bool = False

    def __init__(
        # pylint: disable=redefined-builtin
        self: "Exchange",
        type: Union[str, ExchangeType] = ExchangeType.FANOUT,
        **data,
    ) -> None:
        if isinstance(type, str) and not isinstance(type, ExchangeType):
            data["type"] = ExchangeType(type)
        else:
            data["type"] = type

        super().__init__(**data)



"""
module mqkit.errors.marshalerror

Defines the MarshalError exception indicating errors during marshalling or unmarshalling data.
"""

from .mqkiterror import MqkitError


class MarshalError(MqkitError):
    """Exception raised for errors during marshalling or unmarshalling data."""



"""
module mqkit.marshal

A module providing serialization and message handling utilities for message queue applications.
"""

__all__ = [
    "codecs",
    "FullyTypedSerializer",
    "ReturnTypeSerializer",
    "Serializer",
    "TypelessSerializer",
]

from . import codecs

from .fullytypedserializer import FullyTypedSerializer
from .returntypeserializer import ReturnTypeSerializer
from .serializer import Serializer
from .typelessserializer import TypelessSerializer



"""
module mqkit.workers.worker

Contains the definition of the Worker abstract base class.
"""

from abc import ABCMeta


# pylint: disable=too-few-public-methods
class Worker(metaclass=ABCMeta):
    """
    class Worker

    Abstract base class for all worker implementations.
    """



"""
module mqkit.errors.workerterminatederror

Defines the WorkerTerminatedError exception for handling unexpected worker terminations.
"""

from .mqkiterror import MqkitError


class WorkerTerminatedError(MqkitError):
    """Raised when a worker process has been terminated unexpectedly."""



"""
module mqkit.workers

A module providing worker and coordinator classes for managing
message processing in various concurrency modes.
"""

__all__ = ["Coordinator", "threaded", "Worker"]

from . import threaded

from .coordinator import Coordinator
from .worker import Worker



"""
module mqkit.messaging.retry.retrycontext

This module defines the RetryContext class, which is used to encapsulate the context of a
failed message when handling retries. The RetryContext includes the connection, the
message, and the exception that caused the failure.
"""

from typing import ClassVar, Optional
from pydantic import BaseModel, ConfigDict

from ...connections import Connection
from ..destination import Destination
from ..queuemessage import QueueMessage


class RetryContext(BaseModel):
    """
    class RetryContext

    This class encapsulates the context of a failed message when handling retries. The
    RetryContext includes the connection, the message, and the exception that caused the failure.
    """

    connection: Connection
    message: QueueMessage
    exception: Exception
    received_queue: str
    dead_letter_destination: Optional[Destination] = None

    model_config: ClassVar[ConfigDict] = ConfigDict(
        arbitrary_types_allowed=True,
    )



"""
module mqkit.marshal.codecs.rawcodec

Defines the RawCodec for handling raw byte payloads.
"""

from .codec import Codec


class RawCodec(Codec):
    """
    class RawCodec

    A codec for handling raw byte payloads without any serialization or deserialization.
    """

    @property
    def content_type(self: "RawCodec") -> str:
        return "application/octet-stream"

    def decode(self: "RawCodec", data: bytes) -> bytes:
        return data

    def encode(self: "RawCodec", data: bytes) -> bytes:
        return data



"""
module mqkit.warnings

This module defines the warnings that are used in the mqkit package. These warnings are used to
inform users of potential issues with their configuration or usage of the package. The warnings
defined in this module include:
    - ConfigurationWarning: A warning that is emitted when a user has a potential issue with their
        configuration.
    - MqkitWarning: Base class for all mqkit warnings
    - UnboundQueueWarning: A warning that is emitted when a user defines a forward queue for an
        endpoint that doesn't have a handler defined.
"""

__all__ = [
    "ConfigurationWarning",
    "MqkitWarning",
    "UnboundQueueWarning",
]

from .configurationwarning import ConfigurationWarning
from .mqkitwarning import MqkitWarning
from .unboundqueuewarning import UnboundQueueWarning



"""
module mqkit.endpoints.queueendpoint

Defines the QueueEndpoint class for processing messages from a message queue.
"""

from copy import copy
from typing import TYPE_CHECKING, Optional, override

from .endpoint import Endpoint
from ..errors import NoForwardTargetError
from ..messaging import Attributes, Destination, Forward, QueueMessage, Response
from ..messaging.retry import RetryStrategy

if TYPE_CHECKING:
    from .config import QueueEndpointConfig


class QueueEndpoint(Endpoint):
    """
    class QueueEndpoint

    Represents an endpoint for processing messages from a message queue.
    """

    _config: "QueueEndpointConfig"

    def __init__(
        # pylint: disable=too-many-arguments,too-many-positional-arguments
        self: "QueueEndpoint",
        config: "QueueEndpointConfig",
    ) -> None:
        super().__init__(
            target=config.target,
            codec_type=config.codec_type,
        )

        self._config = config

    @property
    def dead_letter(self: "QueueEndpoint") -> Optional[Destination]:
        return self._config.dead_letter

    @property
    def forward_target(self: "QueueEndpoint") -> Optional[Destination]:
        """
        Get the forward target destination for this endpoint.

        Returns:
            Optional[Destination]: The forward target destination if configured, otherwise None.
        """

        return self._config.forward_to

    def _forward_result(self: "QueueEndpoint", response: Response) -> Optional[Forward]:
        assert self._config.forward_to is not None

        if not response.has_data:  # pragma: no cover
            raise ValueError("Cannot forward response with no data")
        if response.topic is not None and self._config.forward_to.topic is not None:
            raise ValueError(
                "Cannot forward response with topic when forward_to destination also has a topic"
            )

        if isinstance(self._config.forward_to, Destination):
            return Forward(
                forward_target=self._config.forward_to.resource,
                message=QueueMessage(
                    data=response.data,
                    attributes=Attributes(
                        headers=self.make_forward_headers(
                            response,
                            origin_queue=self._config.queue.name,
                        ),
                        forwarded=True,
                        origin_queue=self._config.queue.name,
                        topic=response.topic or self._config.forward_to.topic,
                        is_dead_letter=False,
                    ),
                ),
            )

        raise NotImplementedError(
            f"Forwarding to targets of type {type(self._config.forward_to).__name__} "
            "is not implemented"
        )  # pragma: no cover

    def handle_message(
        self: "QueueEndpoint", message: QueueMessage
    ) -> Optional[Forward]:
        """
        Handle an incoming message by invoking the target function.

        Args:
            message (QueueMessage): The incoming message to handle.

        Returns:
            Optional[Forward]: A Forward object if the result should be forwarded,
                otherwise None.

        Raises:
            ValueError: If the target function returns a result but no forward_to
                queue was specified.
        """

        result: Optional[Response]
        try:
            result = self.target(
                message=message.data,
                attributes=message.attributes,
            )
        except Exception as exc:
            # if this exception type has a handler, call it as a side effect
            if type(exc) in self._config.error_handlers:
                self._config.error_handlers[type(exc)](  # type: ignore
                    message.data,
                    message.attributes,
                    exc,
                )

            raise exc

        if result is None:
            return None
        if not isinstance(result, Response):  # pragma: no cover
            raise TypeError(
                f"Target function returned invalid type {type(result).__name__}; "
                "expected Response or None"
            )

        # check if the message can actually be replied to
        if self._config.forward_to is None:
            raise NoForwardTargetError(
                "Cannot forward returned message result because no forward_to queue was specified"
            )

        return self._forward_result(result)

    @property
    @override
    def is_auto_delete(self: "QueueEndpoint") -> bool:
        return self._config.queue.auto_delete

    @property
    @override
    def is_persistent(self: "QueueEndpoint") -> bool:
        return self._config.queue.persistent

    @property
    def qualname(self: "QueueEndpoint") -> str:  # pragma: no cover
        return self.queue_name

    @property
    def queue_name(self: "QueueEndpoint") -> str:
        return copy(self._config.queue.name)

    @property
    def retry_strategy(self: "QueueEndpoint") -> RetryStrategy:
        return self._config.retry_strategy



"""
module mqkit.connections.amqp.amqpsentinel

Defines the AmqpSentinel exception used as a sentinel value in AMQP connections.
An instance of this exception can be passed into the internal Queue to signal
that the connection should be shut down
"""

from ...errors import MqkitError


class AmqpSentinel(MqkitError):
    """
    Sentinel exception used to signal shutdown in AMQP connections.
    """



"""
module mqkit.endpoints.endpointfactory

Defines the EndpointFactory class for creating endpoint instances.
"""

from typing import TYPE_CHECKING, Optional

from ..messaging import Destination
from .queueendpoint import QueueEndpoint

if TYPE_CHECKING:
    from .config import QueueEndpointConfig
    from ..messaging import ForwardTarget


class EndpointFactory:
    """
    class EndpointFactory

    Factory class for creating endpoint instances.
    """

    # pylint: disable=too-few-public-methods

    @staticmethod
    def create_queue_endpoint(config: "QueueEndpointConfig") -> QueueEndpoint:
        """
        Creates a QueueEndpoint instance based on the provided configuration.

        Args:
            config (QueueEndpointConfig): The configuration for the queue endpoint.

        Returns:
            QueueEndpoint: The created QueueEndpoint instance.
        """

        return QueueEndpoint(config)

    @staticmethod
    def convert_forward_target_to_destination(
        forward_to: Optional["ForwardTarget"] = None,
    ) -> Optional[Destination]:
        """
        Converts a forward target to a Destination object.

        Args:
            forward_to (Optional[Union[str, Queue, Exchange, Destination]]): The forward target
                to convert.

        Returns:
            Optional[Destination]: The converted Destination object, or None if the input is None.
        """

        return Destination.from_forward_target(forward_to)



"""
module mqkit.workers.threaded.threadcoordinator

Defines the ThreadCoordinator class for managing threaded workers
to process messages from multiple endpoints concurrently.
"""

from logging import Logger
import logging
from queue import Queue as ProcessQueue, ShutDown
from typing import Dict, List, Union

from ..coordinator import Coordinator
from ...endpoints import Endpoint
from ...engines import Engine
from ...errors import FunctionTypeError, ShutdownRequested, WorkerTerminatedError
from ...logging import root_logger_name
from .monotoniccounter import MonotonicCounter
from .threadworker import ThreadWorker


class ThreadCoordinator(Coordinator):
    """
    class ThreadCoordinator

    A coordinator that manages workers for processing messages from
    multiple endpoints concurrently using threading.
    """

    # pylint: disable=too-few-public-methods

    _allow_restart: bool
    _error_queue: ProcessQueue = ProcessQueue()
    _logger: Logger
    _restart_counter: MonotonicCounter
    _serial_number_counter: MonotonicCounter = MonotonicCounter()
    _serial_number: int
    _workers: Dict[ThreadWorker, Endpoint]

    def __init__(
        self: "ThreadCoordinator",
        endpoints: List[Endpoint],
        engine: Engine,
        allow_restart: bool = False,
    ) -> None:
        super().__init__(endpoints, engine)

        # assert that all endpoints are compatible with threaded concurrency
        self._assert_endpoints_compatible()
        self._serial_number = self._serial_number_counter.next()
        self._logger = logging.getLogger(
            f"{root_logger_name}.{'.'.join(self.__class__.__module__.split('.')[1:-1])}."
            f"{self.__class__.__name__}-{self._serial_number}"
        )

        self._allow_restart = allow_restart
        self._error_queue = ProcessQueue()
        self._restart_counter = MonotonicCounter()

    def _assert_endpoints_compatible(self: "ThreadCoordinator") -> None:
        if any(endpoint.is_async for endpoint in self._endpoints):
            raise FunctionTypeError(
                "All endpoints must be compatible with threaded concurrency (async not allowed)"
            )

    def _handle_errors(self: "ThreadCoordinator") -> None:
        while not self._error_queue.is_shutdown:
            error_worker: ThreadWorker
            try:
                error_worker = self._error_queue.get()
            except ShutDown:
                continue

            self._logger.error(
                "Worker %s for endpoint '%s' exited with exception: %s",
                error_worker.name,
                self._workers[error_worker].qualname,
                error_worker.error,
            )

            if not self._allow_restart:
                # if we're not allowed to restart workers, raise an error
                raise WorkerTerminatedError(
                    f"{error_worker.name} terminated and cannot be restarted by coordinator "
                    "(restarts not allowed)"
                ) from error_worker.error

            self._restart_worker(error_worker)

    def _interrupt(
        self: "ThreadCoordinator",
        exception: Union[Exception, KeyboardInterrupt],
    ) -> None:
        self._error_queue.shutdown()
        self._stop_workers(
            exception=exception,
            reason=(
                "User interaction"
                if isinstance(exception, KeyboardInterrupt)
                else str(exception)
            ),
        )

    def _restart_worker(
        self: "ThreadCoordinator",
        worker: ThreadWorker,
    ) -> None:
        endpoint: Endpoint = self._workers[worker]
        self._logger.info(
            "Restarting worker for endpoint '%s' (restart #%d)",
            endpoint.qualname,
            self._restart_counter.next(),
        )

        new_worker: ThreadWorker = ThreadWorker(
            endpoint=endpoint,
            engine=self._engine,
            error_queue=self._error_queue,
        )
        self._workers[new_worker] = endpoint
        del self._workers[worker]

        new_worker.start()
        self._logger.info(
            "Worker for endpoint '%s' restarted successfully",
            endpoint.qualname,
        )

    def _start_workers(self: "ThreadCoordinator") -> None:
        self._logger.debug(
            "Starting ThreadCoordinator with %d worker%s",
            len(self._workers),
            "s" if len(self._workers) != 1 else "",
        )
        for worker in self._workers:
            worker.start()

        self._logger.debug("All workers started")
        self._handle_errors()

    def _stop_workers(
        self: "ThreadCoordinator",
        exception: Union[Exception, KeyboardInterrupt],
        reason: str,
    ) -> None:
        self._logger.warning("KeyboardInterrupt received, stopping workers")

        if not hasattr(self, "_workers"):
            self._logger.warning("No workers to stop")
            return

        for worker in self._workers:
            worker.stop(message=f"{reason} ({exception!r})")
        for worker in self._workers:
            worker.join()

        self._logger.warning("All workers stopped")

    def run(self: "ThreadCoordinator") -> None:
        """
        Run the ThreadCoordinator by starting and managing ThreadWorkers for each
        endpoint. Handles graceful shutdown on user interruption via KeyboardInterrupt.

        Args:
            None

        Returns:
            None

        Raises:
            Nothing
        """

        self._workers = {
            ThreadWorker(
                endpoint=endpoint,
                engine=self._engine,
                error_queue=self._error_queue,
            ): endpoint
            for endpoint in self._endpoints
        }

        try:
            self._start_workers()
        except KeyboardInterrupt as ki:  # pragma: no cover
            self._interrupt(ki)
        except WorkerTerminatedError as wti:
            self._interrupt(wti)
            raise

    def stop(
        self: "ThreadCoordinator",
    ) -> None:
        self._interrupt(
            ShutdownRequested(
                "Shutdown requested by application",
            )
        )



"""
module mqkit.apps.concurrencymode

Defines the ConcurrencyMode enumeration for specifying concurrency modes
in message queue applications.
"""

from enum import Enum


class ConcurrencyMode(str, Enum):
    """
    class ConcurrencyMode

    An enumeration representing the concurrency modes available for message
    queue applications.
    """

    THREAD = "thread"
    PROCESS = "process"
    ASYNC = "async"



"""
module mqkit.credentials

Defines the Credentials and PlainCredentials classes for handling
authentication credentials in the mqkit library.
"""

__all__ = [
    "Credentials",
    "PlainCredentials",
]

from .credentials import Credentials
from .plaincredentials import PlainCredentials



"""
module mqkit.connections.connection

Defines the abstract base class Connection for message queue connections.
"""

from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Optional, List, Type

if TYPE_CHECKING:
    from ..declarations import Declaration
    from ..messaging import Forward, QueueMessage


class Connection(metaclass=ABCMeta):
    """
    class Connection

    Abstract base class for message queue connections.
    """

    @abstractmethod
    def acknowledge_failure(self: "Connection", message: "QueueMessage") -> None:
        """
        Acknowledge the failure of processing a message.

        Args:
            message (QueueMessage): The message to acknowledge.

        Returns:
            None

        Raises:
            NotImplementedError: If the method is not implemented.
        """

        raise NotImplementedError()

    @abstractmethod
    def acknowledge_success(self: "Connection", message: "QueueMessage") -> None:
        """
        Acknowledge the processing of a message.

        Args:
            message (QueueMessage): The message to acknowledge.

        Returns:
            None

        Raises:
            NotImplementedError: If the method is not implemented.
        """

        raise NotImplementedError()

    @abstractmethod
    def declare_resources(self: "Connection", resources: List["Declaration"]) -> None:
        """
        Declare a messaging resource (e.g., queue, exchange).

        Args:
            resource (Declaration): The resource to declare.

        Returns:
            None

        Raises:
            NotImplementedError: If the method is not implemented.
        """

        raise NotImplementedError()

    @abstractmethod
    def __enter__(self: "Connection") -> "Connection":
        """
        Enter the runtime context related to this object.

        Args:
            None

        Returns:
            Connection: The connection object itself.

        Raises:
            NotImplementedError: If the method is not implemented.
        """

        raise NotImplementedError()

    @abstractmethod
    def __exit__(
        self: "Connection", exc_type: Type[Exception], exc_value: Exception, traceback
    ) -> None:
        """
        Exit the runtime context related to this object.

        Args:
            exc_type: The exception type.
            exc_value: The exception value.
            traceback: The traceback object.

        Returns:
            None

        Raises:
            NotImplementedError: If the method is not implemented.
        """

        raise NotImplementedError()

    @abstractmethod
    def forward_message(self: "Connection", forward: "Forward") -> None:
        """
        Forward a message to another queue.

        Args:
            forward (Forward): The forward object containing the target and message.

        Returns:
            None

        Raises:
            NotImplementedError: If the method is not implemented.
        """

        raise NotImplementedError()

    @abstractmethod
    def get_message(self: "Connection") -> "QueueMessage":
        """
        Block and wait for a message from the connected queue.

        Args:
            None

        Returns:
            bytes: The message received from the queue.

        Raises:
            NotImplementedError: If the method is not implemented.
        """

        raise NotImplementedError()

    @abstractmethod
    def submit_message(self: "Connection", message: "QueueMessage") -> None:
        """
        Submit a message to the connected queue.

        Args:
            message (QueueMessage): The message to submit.

        Returns:
            None

        Raises:
            NotImplementedError: If the method is not implemented.
        """

        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def unblock(self: "Connection", message: Optional[str] = None) -> None:
        """
        Unblock the connection if it is blocked waiting for a message by
        submitting a sentinel message.

        Args:
            None

        Returns:
            None

        Raises:
            NotImplementedError: If the method is not implemented.
        """

        raise NotImplementedError()



"""
module mqkit.workers.threaded

A module providing threaded worker implementations for message queue processing.
"""

__all__ = [
    "MonotonicCounter",
    "ThreadCoordinator",
    "ThreadWorker",
]

from .monotoniccounter import MonotonicCounter
from .threadcoordinator import ThreadCoordinator
from .threadworker import ThreadWorker



"""
module mqkit.engines.engine

Defines the abstract Engine class for message queue engines. Engines are responsible
for creating connections to message queues.
"""

from abc import ABCMeta, abstractmethod
from typing import List, Type

from pydantic import BaseModel
from yarl import URL

from ..connections import Connection
from ..declarations import Declaration


class Engine(BaseModel, metaclass=ABCMeta):
    """
    class Engine

    An abstract base class for message queue engines. Engines are responsible for
    creating connections to message queues.
    """

    @abstractmethod
    def connect(
        self: "Engine",
        queue: str,
        persistent: bool = True,
        auto_delete: bool = False,
    ) -> Connection:
        """
        Create a connection to the specified message queue. Returns a Connection object
        of the appropriate type for the engine (e.g., AmqpConnection for AMQP engines)

        Args:
            queue (str): The name of the message queue to connect to.

        Returns:
            Connection: A connection object for the specified message queue.

        Raises:
            NotImplementedError: If the method is not implemented.
        """

        raise NotImplementedError()

    @abstractmethod
    def declare_resources(self: "Engine", resources: List[Declaration]) -> None:
        """
        Declare the necessary resources (exchanges, queues, bindings) for the engine.

        Args:
            resources (ExchangeDeclaration): The resources to declare.

        Returns:
            None

        Raises:
            NotImplementedError: If the method is not implemented.
        """

        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def from_url(cls: Type["Engine"], url: URL) -> "Engine":
        """
        Create an engine instance from a URL.

        Args:
            url (URL): The URL to create the engine from.

        Returns:
            Engine: An instance of the engine.

        Raises:
            NotImplementedError: If the method is not implemented.
        """

        raise NotImplementedError()



"""
module mqkit.connections.amqp.amqpconnection

Defines the AmqpConnection class for managing AMQP connections in message
queue applications. This connection class is intended to be returned by
the RabbitMqEngine class
"""

import functools
import json
from logging import Logger
import logging
from queue import Queue as ProcessQueue, ShutDown
import threading
from typing import Any, ClassVar, Dict, List, Optional, Set

from pika import BasicProperties, ConnectionParameters, PlainCredentials, SSLOptions
from pika import BlockingConnection as PikaBlockingConnection
from pika.adapters.blocking_connection import BlockingChannel
from pika.amqp_object import Method
from pydantic import BaseModel, ConfigDict, PrivateAttr, ValidationError
from slugify import slugify

from .amqpconsumethread import AmqpConsumeThread
from .amqpmessage import AmqpMessage
from .amqpsentinel import AmqpSentinel
from ..connection import Connection
from ...declarations import Declaration, ExchangeDeclaration, QueueDeclaration
from ...errors import ShutdownRequested
from ...logging import root_logger_name
from ...messaging import (
    Attributes,
    Exchange,
    ExchangeType,
    Forward,
    Queue,
    QueueMessage,
)
from ...messaging.exceptionhistoryentry import ExceptionHistoryEntry


class AmqpConnection(Connection, BaseModel):
    """
    class AmqpConnection

    Manages an AMQP connection using Pika's BlockingConnection. Provides methods
    for acknowledging message processing results, forwarding messages, and
    retrieving messages from the queue
    """

    host: str
    port: int
    vhost: str
    credentials: PlainCredentials
    queue: Queue
    use_ssl: bool = False

    _channel: Optional[BlockingChannel] = PrivateAttr(default=None)
    _connection: Optional[PikaBlockingConnection] = PrivateAttr(default=None)
    _consume_thread: Optional[AmqpConsumeThread] = PrivateAttr(default=None)
    _declared_exchanges: Set[str] = PrivateAttr(default_factory=set)
    _declared_queues: Set[str] = PrivateAttr(default_factory=set)
    _logger: Logger = PrivateAttr()
    _message_queue: ProcessQueue = PrivateAttr(default_factory=ProcessQueue)

    model_config: ClassVar[ConfigDict] = ConfigDict(arbitrary_types_allowed=True)

    def model_post_init(self: "AmqpConnection", context: Any) -> None:  # pylint: disable=arguments-differ
        self._logger = logging.getLogger(
            f"{root_logger_name}.{'.'.join(self.__class__.__module__.split('.')[1:-1])}."
            f"{self.__class__.__name__}.{self.queue}"
        )

    def acknowledge_failure(
        self: "AmqpConnection",
        message: QueueMessage,
    ) -> None:
        if self._connection is None or self._channel is None:
            raise RuntimeError("AMQP channel is not established")

        self._connection.add_callback_threadsafe(
            functools.partial(
                self._channel.basic_nack,
                delivery_tag=self._get_delivery_tag(message),
                requeue=False,
            )
        )

    def acknowledge_success(self: "AmqpConnection", message: QueueMessage) -> None:
        if self._connection is None or self._channel is None:
            raise RuntimeError("AMQP channel is not established")

        self._connection.add_callback_threadsafe(
            functools.partial(
                self._channel.basic_ack,
                delivery_tag=self._get_delivery_tag(message),
            )
        )

    def _declare_exchange(
        self: "AmqpConnection",
        exchange: Exchange,
        thread_local: bool = False,
    ) -> None:
        if self._connection is None or self._channel is None:  # pragma: no cover
            raise RuntimeError("AMQP channel is not established")

        # if we've already declared this exchange, skip it
        if exchange.name in self._declared_exchanges:  # pragma: no cover
            return

        done_event: threading.Event = threading.Event()
        declare_exception: Optional[Exception] = None

        def declare() -> None:
            if self._channel is None:  # pragma: no cover
                raise RuntimeError("AMQP channel is not established")

            # pylint: disable=broad-exception-caught
            try:
                self._channel.exchange_declare(
                    exchange=exchange.name,
                    exchange_type=exchange.type.value,
                    durable=exchange.persistent,
                    auto_delete=exchange.auto_delete,
                )
            except Exception as exc:  # pragma: no cover
                nonlocal declare_exception
                declare_exception = exc
            finally:
                done_event.set()

        if not thread_local:  # pragma: no cover
            self._connection.add_callback_threadsafe(declare)
            done_event.wait()
        else:
            declare()

        if declare_exception is not None:  # pragma: no cover
            raise declare_exception

        self._declared_exchanges.add(exchange.name)

    def _declare_queue(
        self: "AmqpConnection",
        queue: Queue,
        thread_local: bool = False,
    ) -> None:
        if self._connection is None or self._channel is None:  # pragma: no cover
            raise RuntimeError("AMQP channel is not established")

        # if we've already declared this queue, skip it
        if queue.name in self._declared_queues:  # pragma: no cover
            return

        done_event: threading.Event = threading.Event()

        def declare() -> None:
            if self._channel is None:  # pragma: no cover
                raise RuntimeError("AMQP channel is not established")

            self._channel.queue_declare(  # pyright: ignore[reportOptionalMemberAccess]
                queue=queue.name,
                durable=queue.persistent,
                auto_delete=queue.auto_delete,
            )
            done_event.set()

        if not thread_local:  # pragma: no cover
            self._connection.add_callback_threadsafe(declare)
            done_event.wait()
        else:
            declare()

        self._declared_queues.add(queue.name)

    def _declare_exchange_with_bindings(
        self: "AmqpConnection",
        exchange_declaration: ExchangeDeclaration,
    ) -> None:
        if self._connection is None or self._channel is None:  # pragma: no cover
            raise RuntimeError("AMQP channel is not established")

        self._declare_exchange(
            exchange_declaration.exchange,
            thread_local=True,
        )

        for binding in exchange_declaration.bindings:
            if isinstance(binding.bound_resource, Queue):
                self._channel.queue_bind(
                    queue=binding.bound_resource.name,
                    exchange=exchange_declaration.exchange.name,
                    routing_key=binding.topic,
                )
                continue
            if isinstance(binding.bound_resource, Exchange):
                self._channel.exchange_bind(
                    destination=binding.bound_resource.name,
                    source=exchange_declaration.exchange.name,
                    routing_key=binding.topic,
                )
                continue

            raise NotImplementedError(  # pragma: no cover
                f"Binding resources of type {type(binding.bound_resource).__name__} "
                "is not implemented"
            )

    def _declare_resubmit_exchange(
        self: "AmqpConnection",
        target_queue: Queue,
        thread_local: bool = False,
    ) -> Exchange:
        if self._connection is None or self._channel is None:  # pragma: no cover
            raise RuntimeError("AMQP channel is not established")

        resubmit_exchange: Exchange = Exchange(
            name=self._get_resubmit_exchange(target_queue.name),
            type=ExchangeType.FANOUT,
            persistent=target_queue.persistent,
        )

        # declare the exchange and bind the queue to it
        self._declare_exchange(
            resubmit_exchange,
            thread_local=thread_local,
        )
        if thread_local:
            self._channel.queue_bind(
                queue=target_queue.name,
                exchange=resubmit_exchange.name,
            )
        else:
            self._connection.add_callback_threadsafe(
                functools.partial(
                    self._channel.queue_bind,
                    queue=target_queue.name,
                    exchange=resubmit_exchange.name,
                )
            )

        # return the exchange object we instantiated
        return resubmit_exchange

    def declare_resources(self: "AmqpConnection", resources: List[Declaration]) -> None:
        if self._connection is None or self._channel is None:
            self._connection = self._make_connection()
            self._channel = self._connection.channel()

        try:
            for resource in resources:
                if isinstance(resource, ExchangeDeclaration):
                    self._declare_exchange_with_bindings(resource)
                    continue
                if isinstance(resource, QueueDeclaration):
                    self._declare_queue(
                        resource.queue,
                        thread_local=True,
                    )
                    continue

                raise NotImplementedError(  # pragma: no cover
                    f"Declaring resources of type {type(resource).__name__} is not implemented"
                )
        finally:
            if self._connection is not None and not self._connection.is_closed:
                self._connection.close()
                self._connection = None
                self._channel = None

    def _decode_exception_history(
        self: "AmqpConnection",
        properties: BasicProperties,
        key: str = "x-mqkit-exception-history",
    ) -> List[ExceptionHistoryEntry]:
        if properties.headers is not None:
            try:
                return [
                    ExceptionHistoryEntry(**object)
                    for object in json.loads(
                        properties.headers.get(key, "[]"),  # type: ignore
                    )
                ]
            except (json.JSONDecodeError, ValidationError) as e:
                self._logger.warning(
                    "Failed to decode exception history from message headers, "
                    "proceeding without empty history (%s: %s)",
                    type(e),
                    e,
                )

        return []

    def _decode_retry_count(
        self: "AmqpConnection",
        properties: BasicProperties,
        key: str = "x-mqkit-retry-count",
    ) -> int:
        if properties.headers and key in properties.headers:
            try:
                return int(properties.headers[key])  # type: ignore
            except (ValueError, TypeError):
                pass

        return 0

    def _enqueue_message(
        self: "AmqpConnection",
        channel: BlockingChannel,
        method: Method,
        properties: BasicProperties,
        body: bytes,
    ) -> None:
        self._message_queue.put(
            AmqpMessage(
                channel=channel,
                method=method,
                properties=properties,
                body=body,
            )
        )

    def __enter__(self: "AmqpConnection") -> "AmqpConnection":
        self._connection = self._make_connection()

        # declare the queue as durable if needed and set prefetch count to 1
        self._channel = self._connection.channel()
        self._declare_queue(
            self.queue,
            thread_local=True,
        )
        self._channel.basic_qos(prefetch_count=1)
        self._channel.basic_consume(
            on_message_callback=self._enqueue_message,
            queue=self.queue.name,
            auto_ack=False,
        )

        # declare a default resubmit exchange for failed messages
        self._declare_resubmit_exchange(
            self.queue,
            thread_local=True,
        )

        self._start_consuming()
        return self

    def __exit__(self: "AmqpConnection", exc_type, exc_value, traceback) -> None:
        self._stop_consuming()

        if self._connection is not None and not self._connection.is_closed:
            self._connection.close()

    def _get_delivery_tag(self: "AmqpConnection", message: QueueMessage) -> int:
        return message.attributes.platform["method"]["delivery_tag"]  # type: ignore

    def forward_message(self: "AmqpConnection", forward: Forward) -> None:
        if self._connection is None or self._channel is None:
            raise RuntimeError("AMQP channel is not established")

        if isinstance(forward.forward_target, Queue):
            self._forward_message_to_queue(forward)
            return

        if isinstance(forward.forward_target, Exchange):
            self._declare_exchange(forward.forward_target)
            self._connection.add_callback_threadsafe(
                functools.partial(
                    self._channel.basic_publish,
                    exchange=forward.forward_target.name,
                    routing_key=forward.message.attributes.topic or "",
                    body=forward.message.data,
                    properties=BasicProperties(
                        headers=self._get_forward_message_headers(forward),
                        delivery_mode=2,  # make message persistent
                    ),
                )
            )
            return

        raise NotImplementedError(
            f"Forwarding to targets of type {type(forward.forward_target).__name__} "
            "is not implemented"
        )  # pragma: no cover

    def _forward_message_to_queue(
        self: "AmqpConnection",
        forward: Forward,
    ) -> None:
        if self._connection is None or self._channel is None:  # pragma: no cover
            raise RuntimeError("AMQP channel is not established")
        if not isinstance(forward.forward_target, Queue):  # pragma: no cover
            raise TypeError("Forward target must be a Queue")

        # if a topic was specified, we cannot forward to a queue directly. ensure that
        # a resumbit exchange exists for the target queue and publish to that instead
        if forward.message.attributes.topic is not None:
            self.forward_message(
                Forward(
                    forward_target=self._declare_resubmit_exchange(
                        forward.forward_target
                    ),
                    message=forward.message,
                )
            )
            return

        # if no topic was specified, just publish to the queue directly using the
        # default exchange

        # NOTE: support forward target durability options later. we can maybe infer
        # from other queue definitions
        self._declare_queue(forward.forward_target)
        self._connection.add_callback_threadsafe(
            functools.partial(
                self._channel.basic_publish,
                exchange="",
                routing_key=forward.forward_target.name,
                body=forward.message.data,
                properties=BasicProperties(
                    headers=self._get_forward_message_headers(forward),
                    delivery_mode=2,  # make message persistent
                ),
            )
        )

    def _get_forward_message_headers(
        self: "AmqpConnection",
        forward: Forward,
    ) -> Dict[str, str]:
        headers: Dict[str, str] = forward.message.attributes.headers.copy()

        if forward.message.attributes.retry_count > 0:
            headers["x-mqkit-retry-count"] = str(forward.message.attributes.retry_count)
        if len(forward.message.attributes.exception_history) > 0:
            headers["x-mqkit-exception-history"] = json.dumps(
                [
                    entry.model_dump()
                    for entry in forward.message.attributes.exception_history
                ]
            )

        return headers

    def get_message(self: "AmqpConnection") -> QueueMessage:
        message: AmqpMessage | AmqpSentinel = self._message_queue.get()

        if isinstance(message, AmqpSentinel):
            self._message_queue.shutdown()
            raise ShutdownRequested(*message.args)

        return QueueMessage(
            data=message.body,
            attributes=Attributes(
                headers={
                    key: str(value)
                    for key, value in (message.properties.headers or {}).items()
                    if not key.startswith("x-mqkit-")
                },
                platform={
                    "channel": {
                        "number": message.channel.channel_number,
                    },
                    "method": {
                        "consumer_tag": message.method.consumer_tag,  # type: ignore
                        "delivery_tag": message.method.delivery_tag,  # type: ignore
                        "exchange": message.method.exchange,  # type: ignore
                        "redelivered": message.method.redelivered,  # type: ignore
                        "routing_key": message.method.routing_key,  # type: ignore
                    },
                    "properties": {
                        "delivery_mode": message.properties.delivery_mode,
                    },
                },
                forwarded=(
                    bool(message.properties.headers.get("x-mqkit-forwarded", "false"))
                    if message.properties.headers
                    else False
                ),
                origin_queue=(
                    str(message.properties.headers.get("x-mqkit-origin-queue"))
                    if message.properties.headers
                    else None
                ),
                topic=(
                    message.method.routing_key  # type: ignore
                    if message.method.exchange != ""  # type: ignore
                    else None
                ),
                retry_count=self._decode_retry_count(message.properties),
                previous_retry_count=self._decode_retry_count(
                    message.properties,
                    key="x-mqkit-previous-retry-count",
                ),
                is_dead_letter=(
                    (
                        str(
                            message.properties.headers.get(
                                "x-mqkit-dead-letter", "false"
                            )
                        ).lower()
                        == "true"
                    )
                    if message.properties.headers
                    else False
                ),
                exception_history=self._decode_exception_history(message.properties),
            ),
        )

    def _make_connection(self: "AmqpConnection") -> PikaBlockingConnection:
        ssl_options: Optional[SSLOptions] = None
        if self.use_ssl:
            raise NotImplementedError(
                "SSL connections are not yet implemented for RabbitMQ"
            )

        return PikaBlockingConnection(
            parameters=ConnectionParameters(
                host=self.host,
                port=self.port,
                virtual_host=self.vhost,
                credentials=self.credentials,
                ssl_options=ssl_options,
            )
        )

    def _get_resubmit_exchange(self: "AmqpConnection", queue_name: str) -> str:
        return f"mqkit.resubmit.{slugify(queue_name, separator='_')}"

    @property
    def resubmit_exchange(self: "AmqpConnection") -> str:  # pragma: no cover
        """
        Property that returns the name of the resubmit exchange for the queue
        associated with this connection.

        Returns:
            str: The name of the resubmit exchange.
        """

        return self._get_resubmit_exchange(self.queue.name)

    def _start_consuming(self: "AmqpConnection") -> None:
        if self._channel is None:
            raise RuntimeError("AMQP channel is not established")

        self._consume_thread = AmqpConsumeThread(
            channel=self._channel,
            daemon=True,
        )
        self._consume_thread.start()  # type: ignore

    def _stop_consuming(self: "AmqpConnection") -> None:
        if self._consume_thread is not None:
            self._consume_thread.stop()
            self._consume_thread.join()

            self._consume_thread = None

    def submit_message(self: "AmqpConnection", message: "QueueMessage") -> None:
        self.forward_message(
            Forward(
                forward_target=self.queue,
                message=message,
            )
        )

    def unblock(self: "AmqpConnection", message: Optional[str] = None) -> None:
        if self._connection is None:
            raise RuntimeError("AMQP connection is not established")

        try:
            self._message_queue.put(AmqpSentinel(message))
        except ShutDown:
            self._logger.warning(
                "Failed to unblock connection because message queue is shut down"
            )



"""
module mqkit.engines

Defines the Engine base class and factory method to create specific engine instances
based on connection URLs.
"""

__all__ = ["create_engine", "Engine", "RabbitMqEngine"]

from typing import Dict, Type

from yarl import URL

from .engine import Engine
from .rabbitmq import RabbitMqEngine


_scheme_mapping: Dict[str, Type[Engine]] = {
    "amqp": RabbitMqEngine,
    "amqps": RabbitMqEngine,
}


def create_engine(url: str) -> Engine:
    """
    Factory method to create an Engine instance based on the provided URL.

    Args:
        url (str): The connection URL for the messaging engine.

    Returns:
        Engine: An instance of the appropriate Engine subclass.
    """

    connect_url: URL = URL(url)
    if connect_url.scheme not in _scheme_mapping:
        raise ValueError(
            f"Unknown URL scheme '{connect_url.scheme}' for engine creation"
        )

    return _scheme_mapping[connect_url.scheme].from_url(connect_url)



"""
module mqkit.messaging.response

Defines the Response class for handling message responses.
"""

from typing import Dict, Generic, Optional, TypeVar

from pydantic import BaseModel, Field, PrivateAttr

T = TypeVar("T")


class Response(BaseModel, Generic[T]):
    """
    class Response

    Represents a response message with data and optional headers.
    """

    content: T
    headers: Dict[str, str] = Field(default_factory=dict)
    topic: Optional[str] = None

    _serialized_data: Optional[bytes] = PrivateAttr(default=None)

    def __init__(
        self: "Response",
        content: T,
        headers: Optional[Dict[str, str]] = None,
        topic: Optional[str] = None,
    ) -> None:
        super().__init__(
            content=content,
            headers=headers or {},
            topic=topic,
        )

    @property
    def data(self: "Response") -> bytes:
        """
        Property data

        Gets the serialized data of the response.

        Returns:
            bytes: The serialized data if set

        Raises:
            ValueError: If the serialized data is not set.
        """

        if self._serialized_data is None:
            raise ValueError(
                "Serialized data is not set (Response has not been serialized)"
            )

        return self._serialized_data

    @data.setter
    def data(self: "Response", value: Optional[bytes]) -> None:
        """
        Property setter data

        Sets the serialized data of the response.

        Args:
            value (bytes): The serialized data to set.
        """

        if not isinstance(value, bytes):
            raise TypeError(
                f"Serialized data must be of type bytes, not {type(value).__name__}"
            )

        self._serialized_data = value

    @property
    def has_data(self: "Response") -> bool:
        """
        Property has_data

        Indicates whether the serialized data is set.

        Returns:
            bool: True if the serialized data is set, False otherwise.
        """

        return self._serialized_data is not None



"""
module mqkit.messaging.destination

Defines the Destination class representing a messaging destination.
"""

from typing import Optional, Union

from pydantic import BaseModel

from .exchange import Exchange
from .queue import Queue


class Destination(BaseModel):
    """
    class Destination

    Represents a messaging destination, which can be either a Queue or an Exchange.
    """

    resource: Union[Queue, Exchange]
    topic: Optional[str] = None

    @classmethod
    def from_forward_target(
        cls,
        forward_to: Optional[Union[str, Queue, Exchange, "Destination"]] = None,
    ) -> Optional["Destination"]:
        """
        Converts a forward target to a Destination object.

        Args:
            forward_to (Optional[Union[str, Queue, Exchange, Destination]]): The forward target
                to convert. Can be a string (queue name), a Queue, an Exchange, or a Destination.

        Returns:
            Optional[Destination]: The converted Destination object, or None if the input is None.
        """

        if forward_to is None:
            return None

        if isinstance(forward_to, str):
            forward_to = Queue(name=forward_to)

        if isinstance(forward_to, (Queue, Exchange)):
            return cls(resource=forward_to)

        return forward_to



"""
module mqkit

A message queue toolkit for building and managing message queue applications.
"""

__all__ = [
    "App",
    "consume",
    "create_engine",
    "Exchange",
    "ImmediateRetryStrategy",
    "NoRetry",
    "NoRetryStrategy",
    "Queue",
    "Response",
    "RetryStrategy",
    #
    "apps",
    "credentials",
    "declarations",
    "endpoints",
    "engines",
    "errors",
    "events",
    "logging",
    "marshal",
    "messaging",
    "warnings",
    "workers",
]

from . import apps
from . import credentials
from . import declarations
from . import endpoints
from . import engines
from . import errors
from . import events
from . import logging
from . import marshal
from . import messaging
from . import warnings
from . import workers

from .messaging.retry import ImmediateRetryStrategy, NoRetryStrategy, RetryStrategy

from . import consume as _consume

consume = _consume.consume
del _consume

App = apps.App
Attributes = messaging.Attributes
create_engine = engines.create_engine
Exchange = messaging.Exchange
NoRetry = errors.NoRetry
Queue = messaging.Queue
Response = messaging.Response



"""
module mqkit.workers.threaded.monotoniccounter

Defines the MonotonicCounter class for generating thread-safe sequential integer values.
"""

import itertools
from threading import Lock


class MonotonicCounter:
    """
    class MonotonicCounter

    A thread-safe monotonic counter that generates sequential integer values.
    Each call to the `next` method returns the next integer in the sequence,
    starting from the specified initial value.
    """

    # pylint: disable=too-few-public-methods

    _counter: itertools.count
    _lock: Lock

    def __init__(self: "MonotonicCounter", start: int = 0) -> None:
        self._counter = itertools.count(start)
        self._lock = Lock()

    def next(self: "MonotonicCounter") -> int:
        """
        Returns the next value in the monotonic counter sequence.

        Returns:
            int: The next value in the sequence.
        """

        with self._lock:
            return next(self._counter)



"""
module mqkit.events.appeventtype

Defines the AppEventType enumeration for application event types that
can result in callbacks being invoked.

The user can register callback functions for these event types using
decorators provided by the App class.
"""

from enum import Enum


class AppEventType(str, Enum):
    """
    class AppEventType

    An enumeration representing the different types of application events
    that can trigger callbacks.
    """

    SHUTDOWN = "shutdown"
    START = "start"



"""
module mqkit.marshal.queue

Defines the Queue model for message queue configurations.
"""

from pydantic import BaseModel


class Queue(BaseModel):
    """
    class Queue

    Model for a message queue configuration.
    """

    name: str
    persistent: bool = True
    auto_delete: bool = False



"""
module mqkit.errors.endpointerror

Defines the EndpointError exception indicating an error related to an endpoint
configuration or operation.
"""

from .mqkiterror import MqkitError


class EndpointError(MqkitError):
    """Exception indicating an error related to an endpoint configuration or operation."""



"""
module mqkit.engines.rabbitmqengine

Defines the RabbitMqEngine class for connecting to RabbitMQ message brokers.
"""

from typing import Any, Dict, List, Type, override
import uuid

from pika import PlainCredentials as PikaPlainCredentials
from yarl import URL

from ...connections.amqp import AmqpConnection
from ...credentials import PlainCredentials
from ...declarations import Declaration
from ..engine import Engine
from ...messaging import Queue


class RabbitMqEngine(Engine):
    """
    class RabbitMqEngine

    An engine for connecting to RabbitMQ message brokers using AMQP protocol.
    Provides methods to create AMQP connections to specified queues.
    """

    host: str
    credentials: PlainCredentials
    port: int = 5672
    vhost: str = "/"
    use_amqps: bool = False

    def connect(
        self: "RabbitMqEngine",
        queue: str,
        persistent: bool = True,
        auto_delete: bool = False,
    ) -> AmqpConnection:
        return self._make_connection(
            queue=queue,
            persistent=persistent,
            auto_delete=auto_delete,
        )

    @override
    def declare_resources(self: "RabbitMqEngine", resources: List[Declaration]) -> None:
        # create an uninitialized connection to perform resource declarations. the uuid
        # queue will never actually get created since __enter__() isn't called
        self._make_connection(
            queue=f"{uuid.uuid4()}",
            persistent=True,
            auto_delete=False,
        ).declare_resources(resources)

    @classmethod
    def from_url(cls: Type["RabbitMqEngine"], url: URL) -> "RabbitMqEngine":
        assert url.scheme in ("amqp", "amqps")

        for required_attribute in ("host", "user", "password"):
            if getattr(url, required_attribute) is None:
                raise AttributeError(
                    f"URL is missing required attribute: {required_attribute}"
                )

        ctor_args: Dict[str, Any] = {
            "host": url.host,
            "credentials": PlainCredentials(
                username=url.user,  # type: ignore
                password=url.password,  # type: ignore
            ),
            "use_amqps": url.scheme == "amqps",
        }
        if url.path not in (None, ""):
            ctor_args["vhost"] = url.path
        if url.port is not None:
            ctor_args["port"] = url.port
        elif url.scheme == "amqps":
            ctor_args["port"] = 5671

        return RabbitMqEngine(**ctor_args)

    def _make_connection(
        self: "RabbitMqEngine",
        queue: str,
        persistent: bool,
        auto_delete: bool,
    ) -> AmqpConnection:
        return AmqpConnection(
            host=self.host,
            port=self.port,
            vhost=self.vhost,
            credentials=PikaPlainCredentials(
                self.credentials.username, self.credentials.password
            ),
            queue=Queue(
                name=queue,
                persistent=persistent,
                auto_delete=auto_delete,
            ),
            use_ssl=self.use_amqps,
        )



"""
module mqkit.connections.amqp

A module providing AMQP connection and consumption thread classes for message queue applications.
"""

__all__ = [
    "AmqpConnection",
    "AmqpConsumeThread",
]

from .amqpconnection import AmqpConnection
from .amqpconsumethread import AmqpConsumeThread



"""
module mqkit.warnings.mqkitwarning

Defines the MqkitWarning class for general warnings related to the mqkit library.
"""


class MqkitWarning(UserWarning):
    """
    class MqkitWarning

    Parent class for all warnings related to the mqkit library. Inherits from UserWarning.
    """



"""
module mqkit.warnings.unboundqueuewarning

This module defines the UnboundQueueWarning class, which is a warning that is emitted when
a user defines a forward queue for an endpoint that doesn't have a handler defined.
"""

from .configurationwarning import ConfigurationWarning


class UnboundQueueWarning(ConfigurationWarning):
    """
    class UnboundQueueWarning

    Warning that is emitted when a user defines a forward queue for an endpoint
    that doesn't have a handler defined
    """



"""
module mqkit.errors.noforwardtargeterror

Defines the NoForwardTargetError exception indicating that a message
"""

from .messageerror import MessageError


class NoForwardTargetError(MessageError):
    """Raised when a message has no forward target defined."""



"""
module mqkit.messaging.exceptionhistoryentry

This module defines the ExceptionHistoryEntry class, which is used to represent an entry in the
exception history of a failed message. The ExceptionHistoryEntry includes the type, module,
message, traceback, and retry count of the exception.
"""

from typing import List
from pydantic import BaseModel


class ExceptionHistoryEntry(BaseModel):
    """
    class ExceptionHistoryEntry

    This class represents an entry in the exception history of a failed message. The
    ExceptionHistoryEntry includes the type, module, message, traceback, and retry count of the
    exception.
    """

    exception_type: str
    exception_module: str
    exception_message: str
    traceback: List[str]
    retry_count: int



"""
module mqkit.marshal.returntypeserializer

Defines the ReturnTypeSerializer class responsible for serializing
and deserializing return types in message queue operations.

This serializer ensures that return values can be correctly transmitted
over the message queue by converting them to and from a suitable format.
It does not enforce any constraints on parameters
"""

import inspect
from types import UnionType
from typing import (
    Any,
    Callable,
    Dict,
    Tuple,
    get_args,
    get_origin,
    Optional,
    Type,
    Union,
)

from pydantic import BaseModel

from mqkit.messaging.response import Response

from ..errors import FunctionSignatureError, SerializeError
from ..marshal.codecs import Codec
from .serializer import Serializer


class ReturnTypeSerializer(Serializer):
    """
    class ReturnTypeSerializer

    Responsible for serializing and deserializing return types in message queue operations.
    This serializer ensures that return values can be correctly transmitted over the message
    queue by converting them to and from a suitable format. It does not enforce any constraints
    on parameters.
    """

    _return_type: Union[None, Type[BaseModel], Type[Dict[str, Any]]]

    def __init__(
        self: "ReturnTypeSerializer", function: Callable, codec: Codec
    ) -> None:
        """
        Initialize the ReturnTypeSerializer with the given function and codec.

        Args:
            function (Callable): The function whose return type is to be serialized.
            codec (Codec): The codec used for serialization and deserialization.

        Returns:
            None
        """

        super().__init__(function, codec)

        # check the validity of the function annotations
        signature: inspect.Signature = inspect.signature(function)
        if "return" not in function.__annotations__:
            raise FunctionSignatureError("Function must have a return type annotation")
        if len(signature.parameters) != 2:
            raise FunctionSignatureError("Function must have exactly two parameters")

        # check that the return type annotation is valid
        self._return_type = signature.return_annotation
        if (
            self._return_type is not None
            and self._return_type is not dict
            and get_origin(signature.return_annotation) is not dict
            and (
                isinstance(signature.return_annotation, type)
                and not issubclass(signature.return_annotation, BaseModel)
            )
        ):
            raise FunctionSignatureError(
                "Return type annotation must be a subclass of BaseModel or Dict"
            )

        self._return_func = self._get_return_func_for_type(
            self._return_type,
            signature=signature,
        )

    def _get_return_func_for_type(
        self: "ReturnTypeSerializer",
        return_type: Any,
        signature: inspect.Signature,
    ) -> Callable[[Any], Any]:
        # check if the return type is Optional
        if get_origin(return_type) in (Union, UnionType):
            return self._get_return_func_for_union_type(
                return_type,
                signature=signature,
            )

        if return_type is None:
            return self._serialize_none
        if isinstance(signature.return_annotation, type) and issubclass(
            return_type, Response
        ):
            return self._get_return_func_for_response_type(return_type)
        if isinstance(signature.return_annotation, type) and issubclass(
            return_type, BaseModel
        ):
            return self._serialize_base_model
        if return_type is dict or get_origin(return_type) is dict:
            return self._serialize_dict

        raise FunctionSignatureError(
            "Return type annotation must be a subclass of BaseModel, Dict, Response, or None"
        )  # pragma: no cover

    def _get_return_func_for_response_type(
        self: "ReturnTypeSerializer",
        return_type: Any,
    ) -> Callable[[Any], Any]:
        type_args: Tuple = (
            return_type.__pydantic_generic_metadata__["args"]
            if hasattr(return_type, "__pydantic_generic_metadata__")
            and "args" in return_type.__pydantic_generic_metadata__
            else tuple()
        )
        if len(type_args) != 1:  # pragma: no cover
            raise FunctionSignatureError(
                "Response return type annotation must have exactly one content type "
                f"({len(type_args)} found)"
            )

        response_content_type: Type = type_args[0]
        if not issubclass(response_content_type, BaseModel):
            raise FunctionSignatureError(
                "Response return type annotation must have a BaseModel content type"
            )

        # extract the BaseModel type from the Response generic
        self._return_type = response_content_type
        return self._serialize_base_model

    def _get_return_func_for_union_type(
        self: "ReturnTypeSerializer",
        return_type: Any,
        signature: inspect.Signature,
    ) -> Callable[[Any], Any]:
        type_args: Tuple = get_args(return_type)

        # Check for Optional[T] (i.e., Union[T, None])
        if len(type_args) == 2 and type_args[1] is type(None):
            return self._make_return_func_optional(
                type_args,
                signature=signature,
            )

        raise FunctionSignatureError(
            f"Return type annotation {return_type} is not supported"
        )  # pragma: no cover

    def _make_return_func_optional(
        self: "ReturnTypeSerializer",
        type_args: Tuple,
        signature: inspect.Signature,
    ) -> Callable[[Any], Any]:
        # first, get the base return func for the non-None type in the Optional
        signature = signature.replace(return_annotation=type_args[0])
        non_none_return_func: Callable = self._get_return_func_for_type(
            type_args[0],
            signature=signature,
        )

        def optional_return_func(data: Any) -> Any:
            if data is None:
                return None

            return non_none_return_func(data)

        return optional_return_func

    @property
    def return_type(self: "ReturnTypeSerializer") -> Any:
        """
        Property that returns the return type annotation of the function.

        Returns:
            Any: The return type annotation.
        """

        return self._return_type

    def _serialize_none(self: "ReturnTypeSerializer", data: Any) -> None:
        if data is not None:
            raise SerializeError(
                "Function return type is None, but returned value is not None "
                f"(type {type(data).__name__})"
            )

    def _serialize_base_model(
        self: "ReturnTypeSerializer", data: Any
    ) -> Dict[str, Any]:
        if not isinstance(data, self._return_type):  # type: ignore
            raise SerializeError(
                "Function return type is "
                f"{self._return_type.__name__}, but returned value is of type "  # type: ignore
                f"{type(data).__name__}"
            )

        return data.model_dump()

    def _serialize_dict(self: "ReturnTypeSerializer", data: Any) -> Dict[str, Any]:
        if not isinstance(data, dict):
            raise SerializeError(
                "Function return type is Dict, but returned value is of type "
                f"{type(data).__name__}"
            )

        return data

    def deserialize(self: "Serializer", data: bytes) -> Any:
        return self._codec.decode(data)

    def serialize(self: "ReturnTypeSerializer", data: Any) -> Optional[bytes]:
        converted_data: Optional[Dict[str, Any]] = self._return_func(data)
        if converted_data is None:
            return None

        return self._codec.encode(converted_data)



"""
module mqkit.errors.configurationerror

This module defines the ConfigurationError class, which represents an error that occurs when
there is a problem with the configuration of an app/consumer. This could include issues such
as invalid parameters, missing required settings, or incompatible configurations. The
ConfigurationError class provides a way to signal that there is a problem with the
configuration and to provide details about the specific issue that occurred.
"""

from .mqkiterror import MqkitError


class ConfigurationError(MqkitError):
    """
    class ConfigurationError

    Represents an error that occurs when there is a problem with the configuration of the system.
    This could include issues such as invalid parameters, missing required settings, or incompatible
    configurations. The ConfigurationError class provides a way to signal that there is a problem
    with the configuration and to provide details about the specific issue that occurred.
    """



"""
module mqkit.apps.app

Contains the definition of the App class for building message queue applications.
"""

import inspect
import logging
from logging import Logger
from typing import Callable, Dict, List, Optional, Set, Union
import warnings

from .concurrencymode import ConcurrencyMode
from ..declarations import Declaration, ExchangeDeclaration, QueueDeclaration
from ..endpoints import Endpoint, EndpointFactory, QueueEndpoint
from ..endpoints.config import QueueEndpointConfig
from ..endpoints.endpoint import EndpointExceptionHandler
from ..engines import Engine
from ..errors import FunctionTypeError
from ..events import AppEventType
from ..logging import root_logger_name
from ..marshal.codecs import CodecType
from ..messaging import Exchange, ExchangeType, ForwardTarget, Queue
from ..messaging.retry import NoRetryStrategy, RetryStrategy
from ..warnings import UnboundQueueWarning
from ..workers import Coordinator
from ..workers.threaded import ThreadCoordinator


class App:
    """
    class App

    Represents a message queue application. Implements a FastAPI-style interface
    for user code to define endpoints and event handlers.
    """

    _codec_type: CodecType
    _concurrency_mode: ConcurrencyMode
    _coordinator: Optional[Coordinator] = None
    _declarations: List[Declaration]
    _endpoints: List[Endpoint] = []
    _event_functions: Dict[AppEventType, Callable] = {}
    _logger: Logger
    _started: bool = False

    def __init__(
        self: "App",
        concurrency_mode: ConcurrencyMode | str = ConcurrencyMode.THREAD,
        codec: CodecType | str = CodecType.JSON,
        logger: Optional[Logger] = None,
    ) -> None:
        self._concurrency_mode = ConcurrencyMode(concurrency_mode)
        self._declarations = []
        self._endpoints = []
        self._codec_type = CodecType(codec)

        self._logger = logger or self._make_default_logger()

    def _assert_function_compatible(
        self: "App",
        func: Callable,
    ) -> None:
        if not self._is_function_compatible(func):
            raise FunctionTypeError(
                ("Async function" if inspect.iscoroutinefunction(func) else "Function")
                + (
                    f" {func.__name__}() is not compatible with concurrency mode "
                    f"'{self.concurrency_mode.value}'"
                )
            )

    @property
    def concurrency_mode(self: "App") -> ConcurrencyMode:
        """
        Property that returns the concurrency mode of the application.

        Returns:
            ConcurrencyMode: The concurrency mode of the application.
        """

        return self._concurrency_mode

    def create_resources(self: "App", engine: Engine) -> None:
        """
        Returns the list of declared resources for the application.

        Args:
            None

        Returns:
            List[Declaration]: The list of declared resources.
        """

        engine.declare_resources(self._declarations)

    def declare(self: "App", resource: Union[Queue, Exchange]) -> Declaration:
        """
        Declares a message queue or exchange resource for the application.

        Args:
            resource (Union[Queue, Exchange]): The message queue or exchange to declare.

        Returns:
            None
        """

        if self._started:
            raise RuntimeError("Cannot declare resources after the app has started")

        declaration: Declaration
        if isinstance(resource, Exchange):
            declaration = ExchangeDeclaration(
                exchange=Exchange(
                    name=resource.name,
                    type=resource.type,
                    persistent=resource.persistent,
                    auto_delete=resource.auto_delete,
                ),
            )
            self._declarations.append(declaration)
            return declaration
        if isinstance(resource, Queue):
            declaration = QueueDeclaration(
                queue=Queue(
                    name=resource.name,
                    persistent=resource.persistent,
                    auto_delete=resource.auto_delete,
                ),
            )
            self._declarations.append(declaration)
            return declaration

        raise TypeError(f"Cannot declare resource of type {type(resource).__name__}")

    def exchange(
        # pylint: disable=redefined-builtin
        self: "App",
        name: str,
        type: Union[ExchangeType, str],
        persistent: bool = True,
        auto_delete: bool = False,
    ) -> ExchangeDeclaration:
        """
        Declares an exchange resource for the application.

        Args:
            name (str): The name of the exchange.
            type (ExchangeType): The type of the exchange.
            persistent (bool): Whether the exchange is persistent. Defaults to True.
            auto_delete (bool): Whether the exchange is auto-delete. Defaults to False.

        Returns:
            None
        """

        declaration: Declaration = self.declare(
            Exchange(
                name=name,
                type=type,
                persistent=persistent,
                auto_delete=auto_delete,
            )
        )
        if not isinstance(declaration, ExchangeDeclaration):  # pragma: no cover
            raise TypeError(
                "Declared exchange did not return ExchangeDeclaration instance"
            )

        return declaration

    def _handle_event(self: "App", event_type: AppEventType) -> None:
        func: Optional[Callable] = self._event_functions.get(event_type)
        if func is not None:
            func()

    def _is_function_compatible(self: "App", func: Callable) -> bool:
        if self.concurrency_mode == ConcurrencyMode.ASYNC:  # pragma: no cover
            return inspect.iscoroutinefunction(func)

        return not inspect.iscoroutinefunction(func)

    @property
    def logger(self: "App") -> Logger:
        """
        Property that returns the logger for the application.

        Returns:
            Logger: The logger for the application.
        """

        return self._logger

    def _make_default_logger(self: "App") -> Logger:
        return logging.getLogger(root_logger_name)

    def on_event(
        self: "App",
        event_type: AppEventType,
    ) -> Callable[[Callable], Callable]:
        """
        Registers a callback function to be called for a specific application event.

        Args:
            event_type (AppEventType): The type of event to register the callback for.
            func (Callable[[], None]): The callback function to register.

        Returns:
            Callable[[], None]: The registered callback function.

        Raises:
            FunctionTypeError: If the function is not compatible with the selected concurrency mode.
        """

        if self._started:
            raise RuntimeError(
                "Cannot register event handlers after the app has started"
            )

        def _on_event_decorator(
            func: Callable[[], None],
        ) -> Callable[[], None]:
            self._assert_function_compatible(func)

            self._event_functions[event_type] = func
            return func

        return _on_event_decorator

    def on_shutdown(
        self: "App",
        func: Callable[[], None],
    ) -> Callable[[], None]:
        """
        Registers a callback function to be called when the application shuts down.

        Args:
            func (Callable[[], None]): The callback function to register.

        Returns:
            Callable[[], None]: The registered callback function.

        Raises:
            FunctionTypeError: If the function is not compatible with the selected concurrency mode.
        """

        return self.on_event(AppEventType.SHUTDOWN)(func)

    def on_start(
        self: "App",
        func: Callable[[], None],
    ) -> Callable[[], None]:
        """
        Registers a callback function to be called when the application starts.

        Args:
            func (Callable[[], None]): The callback function to register.

        Returns:
            Callable[[], None]: The registered callback function.

        Raises:
            FunctionTypeError: If the function is not compatible with the selected concurrency mode.
        """

        return self.on_event(AppEventType.START)(func)

    def queue(
        # pylint: disable=too-many-arguments,too-many-positional-arguments
        self: "App",
        name: str,
        *,
        codec: Optional[CodecType | str] = None,
        forward_to: Optional[ForwardTarget] = None,
        persistent: bool = True,
        auto_delete: bool = False,
        retry_strategy: Optional[RetryStrategy] = None,
        dead_letter: Optional[ForwardTarget] = None,
        on_decode_error: Optional[EndpointExceptionHandler] = None,
        on_validation_error: Optional[EndpointExceptionHandler] = None,
    ) -> Callable[[Callable], QueueEndpoint]:
        """
        Decorator to register a function as a queue endpoint.

        Args:
            name (str): The name of the queue.
            codec (CodecType | str): The codec type for message serialization.
                Defaults to CodecType.JSON.
            forward_to (Optional[str]): The name of the queue to forward results to.
                Defaults to None.

        Returns:
            Callable[[Callable], QueueEndpoint]: The decorator function.

        Raises:
            TypeError: If the function is not compatible with the selected concurrency mode.
        """

        if self._started:
            raise RuntimeError("Cannot register endpoints after the app has started")

        codec = CodecType(codec) if codec is not None else self._codec_type

        # if the user didn't provide an explicit retry strategy, use the default
        # strategy that never performs retries
        if retry_strategy is None:
            retry_strategy = NoRetryStrategy()

        def _queue_decorator(func: Callable) -> QueueEndpoint:
            # check that the function is compatible with the selected concurrency mode
            self._assert_function_compatible(func)

            self._endpoints.append(
                EndpointFactory.create_queue_endpoint(
                    QueueEndpointConfig(
                        queue=Queue(
                            name=name,
                            persistent=persistent,
                            auto_delete=auto_delete,
                        ),
                        target=func,
                        codec_type=codec,
                        forward_to=forward_to,
                        retry_strategy=retry_strategy,
                        dead_letter=dead_letter,
                        error_handlers=QueueEndpointConfig.make_error_handlers_dict(
                            on_decode_error=on_decode_error,
                            on_validation_error=on_validation_error,
                        ),
                    )
                )
            )
            return self._endpoints[-1]  # type: ignore

        return _queue_decorator

    def run(self: "App", engine: Engine) -> None:
        """
        Runs the application using the specified message queue connection engine.

        Args:
            engine (Engine): The message queue connection engine to use.

        Returns:
            None

        Raises:
            RuntimeError: If the application is already running.
            NotImplementedError: If the selected concurrency mode is not implemented.
        """

        self._started = True

        # ensure the app is not already running
        if self._coordinator is not None:
            raise RuntimeError("App is already running")

        # instantiate the appropriate coordinator based on the concurrency mode
        if self.concurrency_mode == ConcurrencyMode.THREAD:
            self._init_threaded(engine)
        else:  # pragma: no cover
            raise NotImplementedError(
                f"Unimplemented concurrency mode '{self.concurrency_mode.value}'"
            )

        # validate all forward targets and emit warnings for any misconfigurations
        # before starting the coordinator
        self._validate_forward_targets()

        # perform any declarations before starting
        self.create_resources(engine)

        # run the coordinator
        assert self._coordinator is not None
        self._handle_event(AppEventType.START)
        self.logger.info(
            "Starting coordinator %s",
            type(self._coordinator).__name__,
        )
        self._coordinator.run()

        # when the coordinator is stopped by an interrupt, trigger the shutdown event
        self.logger.info("Coordinator has stopped, sending shutdown event")
        self._handle_event(AppEventType.SHUTDOWN)

    def _init_threaded(self: "App", engine: Engine) -> None:
        assert self.concurrency_mode == ConcurrencyMode.THREAD

        if self._coordinator is not None:  # pragma: no cover
            raise RuntimeError("App is already running")

        self._coordinator = ThreadCoordinator(
            endpoints=self._endpoints,
            engine=engine,
        )

    @property
    def started(self: "App") -> bool:
        """
        Property that indicates whether the application has been started.

        Returns:
            bool: True if the application has been started, False otherwise.
        """

        return self._started

    def stop(self: "App") -> None:
        """
        Stops the application by stopping the coordinator.

        Args:
            None

        Returns:
            None

        Raises:
            RuntimeError: If the application is not running.
        """

        if self._coordinator is None:
            raise RuntimeError("App is not running")

        self._coordinator.stop()

    def _validate_forward_targets(self: "App") -> None:
        # derive a set of names for queue forward targets and for queues that correspond
        # to a handler
        forward_destination_queue_names: Dict[Endpoint, str] = {}
        handler_queue_names: Set[str] = set()

        for endpoint in self._endpoints:
            # ignore any endpoints that aren't pointing at a queue
            if not isinstance(endpoint, QueueEndpoint):  # pragma: no cover
                raise NotImplementedError(
                    f"{type(endpoint)} not implemented for forward target validation"
                )
            handler_queue_names.add(endpoint.queue_name)

            # don't try getting info from endpoints that don't have a forward target that's a queue
            if endpoint.forward_target is None or not isinstance(
                endpoint.forward_target.resource, Queue
            ):
                continue
            forward_destination_queue_names[endpoint] = (
                endpoint.forward_target.resource.name
            )

        for (
            source_endpoint,
            target_queue_name,
        ) in forward_destination_queue_names.items():
            if target_queue_name not in handler_queue_names:
                warnings.warn(
                    f"Endpoint {source_endpoint.target.__qualname__}() forwards to queue "
                    f"'{target_queue_name}' but no handler is registered for that queue",
                    UnboundQueueWarning,
                    stacklevel=2,
                )



"""
module mqkit.marshal.queuemessage

Defines the QueueMessage data model representing a message in a message queue.
"""

from pydantic import BaseModel

from .attributes import Attributes


class QueueMessage(BaseModel):
    """
    class QueueMessage

    Represents a message in a message queue, including its data and attributes.
    """

    data: bytes
    attributes: Attributes



"""
module mqkit.declarations.queuedeclaration

Defines the QueueDeclaration model representing a queue declaration.
"""

from pydantic import BaseModel

from ..messaging import Queue


class QueueDeclaration(BaseModel):
    """
    class QueueDeclaration

    Model representing a queue declaration.
    """

    queue: Queue



"""
module mqkit.errors.serializeerror

Defines the SerializeError exception indicating an error during serialization
of a message.
"""

from .marshalerror import MarshalError


class SerializeError(MarshalError):
    """Raised when an error occurs during serialization of a message."""



"""
module mqkit.messaging.retry.immediateretrystrategy

This module defines the ImmediateRetryStrategy class, which implements a retry strategy that
immediately retries failed messages up to a specified number of times. If the message continues
to fail after exceeding the maximum retries, it can be forwarded to an optional dead letter
destination for further analysis or handling.
"""

import traceback
from typing import override

from ...errors import ConfigurationError, MarshalError
from ..exceptionhistoryentry import ExceptionHistoryEntry
from ..forward import Forward
from .retrycontext import RetryContext
from .retrystrategy import RetryStrategy


class ImmediateRetryStrategy(RetryStrategy):
    """
    class ImmediateRetryStrategy

    Implements a retry strategy that immediately retries failed messages up to a specified number
    of times. If the message continues to fail after exceeding the maximum retries, it can be
    forwarded to an optional dead letter destination for further analysis or handling.
    """

    # pylint: disable=too-few-public-methods

    _retries: int

    def __init__(
        self: "ImmediateRetryStrategy",
        retries: int,
    ) -> None:
        super().__init__()

        if retries < 0:
            raise ConfigurationError(
                "Retries must be a non-negative integer",
            )

        self._retries = retries

    @override
    def handle_failure(self: "ImmediateRetryStrategy", context: RetryContext) -> None:
        # first, check if the failure was due to a marshalling error. if so, we assume
        # the message was malformed and don't try to retry it
        if issubclass(type(context.exception), MarshalError):
            self._handle_failure_bad_message(context)
            return

        # then, check if we have already exceeded the maximum number of retries. if so,
        # acknowledge failure and do not retry
        if context.message.attributes.retry_count >= self._retries:
            self._handle_failure_retries_exceeded(context)
            return

        # otherwise, submit the message for retry in the origin queue. then, nack the
        # original message to remove it from the queue since we have requeued it
        self._submit_for_retry(context)
        context.connection.acknowledge_failure(context.message)

    def _append_exception_to_history(
        self: "ImmediateRetryStrategy", context: RetryContext
    ) -> None:
        # append an object representing the current exception to the history
        context.message.attributes.exception_history.append(
            ExceptionHistoryEntry(
                exception_type=type(context.exception).__qualname__,
                exception_module=type(context.exception).__module__,
                exception_message=str(context.exception),
                traceback=traceback.format_exception(context.exception),
                retry_count=context.message.attributes.retry_count,
            )
        )

    def _forward_to_dlq(self: "ImmediateRetryStrategy", context: RetryContext) -> None:
        if context.dead_letter_destination is None:
            self._logger.warning(
                "No dead letter destination configured, message will be discarded"
            )
        else:
            self._logger.info(
                "Forwarding failed message to dead letter destination: %s",
                context.dead_letter_destination,
            )

            # set up the message with the appropriate dlq context
            if context.dead_letter_destination.topic is not None:
                context.message.attributes.topic = context.dead_letter_destination.topic
            context.message.attributes.headers["x-mqkit-previous-retry-count"] = str(
                context.message.attributes.retry_count
            )
            context.message.attributes.headers |= {
                "x-mqkit-forwarded": "true",
                "x-mqkit-dead-letter": "true",
                "x-mqkit-origin-queue": context.received_queue,
            }
            self._append_exception_to_history(context)
            context.message.attributes.retry_count = 0

            context.connection.forward_message(
                Forward(
                    forward_target=context.dead_letter_destination.resource,
                    message=context.message,
                )
            )

    def _handle_failure_bad_message(
        self: "ImmediateRetryStrategy", context: RetryContext
    ) -> None:
        self._logger.error(
            "Message handling failed with marshal error: %s. "
            "Will not retry due to malformed message data",
            context.exception,
        )
        self._forward_to_dlq(context)
        context.connection.acknowledge_failure(context.message)

    def _handle_failure_retries_exceeded(
        self: "ImmediateRetryStrategy", context: RetryContext
    ) -> None:
        self._logger.info(
            "Message handling failed with exception: %s. "
            "Exceeded maximum retry count of %s (current retry count: %s). "
            "Acknowledging failure and will not retry",
            context.exception,
            self._retries,
            context.message.attributes.retry_count,
        )

        # forward a copy of the message to the dead letter destination if configured
        # then acknowledge failure on the original message to remove it from the queue
        self._forward_to_dlq(context)
        context.connection.acknowledge_failure(context.message)

    def _submit_for_retry(
        self: "ImmediateRetryStrategy", context: RetryContext
    ) -> None:
        # submit the message for retry by requeuing it with an incremented retry count
        # in the headers
        self._append_exception_to_history(context)
        context.message.attributes.retry_count += 1
        context.connection.submit_message(context.message)
        self._logger.info(
            "Requeued message for immediate retry (current retry count: %s)",
            context.message.attributes.retry_count,
        )



"""
module mqkit.messaging.retry.retrystrategy

This module defines the base class for retry strategies. A retry strategy is responsible for
handling failed messages and determining whether to retry them or not. Subclasses of
RetryStrategy can implement different retry policies
"""

from abc import ABCMeta, abstractmethod
from logging import Logger
import logging

from ...logging import root_logger_name
from .retrycontext import RetryContext


class RetryStrategy(metaclass=ABCMeta):
    """
    class RetryStrategy

    This is the base class for retry strategies. A retry strategy is responsible for handling
    failed messages and determining whether to retry them or not. Subclasses of RetryStrategy
    can implement different retry policies.
    """

    # pylint: disable=too-few-public-methods

    _logger: Logger

    def __init__(self: "RetryStrategy") -> None:
        self._logger = logging.getLogger(
            f"{root_logger_name}.{'.'.join(self.__class__.__module__.split('.')[1:-1])}."
            f"{self.__class__.__name__}"
        )

    @abstractmethod
    def handle_failure(
        self: "RetryStrategy", context: RetryContext
    ) -> None:  # pragma: no cover
        """
        This method is called when a message handler fails to process a message. The retry strategy
        is responsible for handling the failure and determining whether to retry the message or not.

        Args:
            context (RetryContext): The context of the failed message, including the message itself,
            the connection, and any relevant metadata.

        Returns:
            None

        Raises:
            Exception: If the retry strategy encounters an error while handling the failure, it
                should raise an exception to indicate that the failure could not be handled.
        """

        raise NotImplementedError("handle_retry must be implemented by subclasses")



"""
module mqkit.marshal.typelessserializer

Defines the TypelessSerializer class for serializing and deserializing typeless data.
"""

from typing import Any, Dict, Optional

from ..errors import SerializeError
from .serializer import Serializer


class TypelessSerializer(Serializer):
    """
    class TypelessSerializer

    A serializer that handles typeless data represented as dictionaries.
    """

    def deserialize(self: "TypelessSerializer", data: bytes) -> Dict[str, Any]:
        return self._codec.decode(data)

    def serialize(
        self: "TypelessSerializer", data: Optional[Dict[str, Any]]
    ) -> Optional[bytes]:
        if data is None:
            return None
        if not isinstance(data, dict):
            raise SerializeError(
                f"TypelessSerializer can only serialize dict types but got {type(data).__name__}"
            )

        return self._codec.encode(data)



"""
module mqkit.apps

A module providing application framework for message queue applications.
Includes the App class API intended for use by users.
"""

__all__ = ["App", "ConcurrencyMode"]

from .app import App
from .concurrencymode import ConcurrencyMode



"""
module mqkit.errors.messageerror

Module defining the MessageError exception for message processing errors.
"""

from .mqkiterror import MqkitError


class MessageError(MqkitError):
    """Exception indicating an error related to message processing."""



"""
module mqkit.messaging

Defines core messaging classes for message queue applications.
"""

__all__ = [
    "Attributes",
    "Destination",
    "ExceptionHistoryEntry",
    "Exchange",
    "ExchangeType",
    "Forward",
    "ForwardTarget",
    "Queue",
    "QueueMessage",
    "Response",
]


from .attributes import Attributes
from .destination import Destination
from .exceptionhistoryentry import ExceptionHistoryEntry
from .exchange import Exchange
from .exchangetype import ExchangeType
from .forward import Forward
from .forwardtarget import ForwardTarget
from .queue import Queue
from .queuemessage import QueueMessage
from .response import Response



"""
module mqkit.errors.decodeerror

Defines the DecodeError exception indicating an error during decoding
of a message.
"""

from .marshalerror import MarshalError


class DecodeError(MarshalError):
    """Raised when an error occurs during decoding of a message."""



"""
module mqkit.errors.shutdownrequested

Defines the ShutdownRequested exception indicating that a shutdown has been requested
for a message queue worker or application.
"""

from .mqkiterror import MqkitError


class ShutdownRequested(MqkitError):
    """Exception indicating that a shutdown has been requested."""



"""
module mqkit.connections.amqp.amqpmessage

Defines the AmqpMessage data model representing an AMQP message with its associated
channel, method, properties, and body.
"""

from typing import ClassVar
from pika import BasicProperties
from pika.adapters.blocking_connection import BlockingChannel
from pika.amqp_object import Method
from pydantic import BaseModel, ConfigDict


class AmqpMessage(BaseModel):
    """
    class AmqpMessage

    Represents an AMQP message with its associated channel, method, properties, and body.
    """

    channel: BlockingChannel
    method: Method
    properties: BasicProperties
    body: bytes

    model_config: ClassVar[ConfigDict] = ConfigDict(arbitrary_types_allowed=True)



"""
module mqkit.marshal.codecs.messagepackcodec

Defines the MessagePackCodec class for encoding and decoding messages using
the MessagePack format.
"""

from typing import Any, Dict

import msgpack  # type: ignore

from .codec import Codec
from ...errors import DecodeError, EncodeError


class MessagePackCodec(Codec):
    """
    class MessagePackCodec

    Codec for encoding and decoding messages using MessagePack format
    """

    @property
    def content_type(self: "MessagePackCodec") -> str:
        return "application/msgpack"

    def encode(self, data: Dict[str, Any]) -> bytes:
        """
        Encodes a message dictionary into MessagePack bytes.

        Args:
            message (Dict[str, Any]): The message to encode.

        Returns:
            bytes: The encoded MessagePack bytes.
        """

        result: bytes | None = self._pack(data)

        if result is None:  # pragma: no cover
            raise DecodeError("Failed to encode data to MessagePack (value was None)")

        return result

    def decode(self, data: bytes) -> Dict[str, Any]:
        """
        Decodes MessagePack bytes into a message dictionary.

        Args:
            data (bytes): The MessagePack bytes to decode.

        Returns:
            Dict[str, Any]: The decoded message.

        Raises:
            DecodeError: If decoding fails.
        """

        result: Any = self._unpack(data)

        if not isinstance(result, dict):
            raise DecodeError("Decoded MessagePack data is not a dictionary")

        return result

    def _pack(self: "MessagePackCodec", data: Any) -> bytes | None:
        try:
            return msgpack.packb(data, use_bin_type=True)
        except (TypeError, ValueError) as e:
            raise EncodeError("Failed to encode data to MessagePack") from e

    def _unpack(self: "MessagePackCodec", data: bytes) -> Any:
        try:
            return msgpack.unpackb(data, raw=False)
        except ValueError as ve:
            raise DecodeError("Failed to decode MessagePack data") from ve



"""
module mqkit.endpoints.config.queueendpointconfig

Defines the QueueEndpointConfig model for configuring queue endpoints.
"""

from typing import ClassVar, Dict, Optional, Type, Union

from pydantic import BaseModel, ConfigDict, ValidationError

from ..endpoint import (
    EndpointCallback,
    EndpointDecodeException,
    EndpointExceptionHandler,
)
from ..endpointfactory import EndpointFactory
from ...errors import DecodeError
from ...marshal.codecs import CodecType
from ...messaging import Destination, ForwardTarget, Queue
from ...messaging.retry import RetryStrategy


class QueueEndpointConfig(BaseModel):
    """
    class QueueEndpointConfig

    Model for configuring a queue endpoint.
    """

    queue: Queue
    target: EndpointCallback
    codec_type: CodecType
    forward_to: Optional[Destination] = None
    dead_letter: Optional[Destination] = None
    retry_strategy: RetryStrategy
    error_handlers: Dict[Type[EndpointDecodeException], EndpointExceptionHandler] = {}

    model_config: ClassVar[ConfigDict] = ConfigDict(
        arbitrary_types_allowed=True,
    )

    def __init__(
        self: "QueueEndpointConfig",
        codec_type: Union[CodecType, str],
        forward_to: Optional[ForwardTarget] = None,
        dead_letter: Optional[ForwardTarget] = None,
        **data,
    ) -> None:  # pragma: no cover
        if isinstance(codec_type, str):
            data["codec_type"] = CodecType(codec_type)

        data["forward_to"] = EndpointFactory.convert_forward_target_to_destination(
            forward_to
        )
        data["dead_letter"] = EndpointFactory.convert_forward_target_to_destination(
            dead_letter
        )

        super().__init__(**data)

    @staticmethod
    def make_error_handlers_dict(
        on_decode_error: Optional[EndpointExceptionHandler] = None,
        on_validation_error: Optional[EndpointExceptionHandler] = None,
    ) -> Dict[Type[EndpointDecodeException], EndpointExceptionHandler]:
        """
        Constructs a dict mapping exception types to error handler methods. Intended to
        provide standard exception type mappings for public facing interfaces like
        app.queue() and @consume()
        """

        error_handlers: Dict[
            Type[EndpointDecodeException], EndpointExceptionHandler
        ] = {}

        if on_decode_error is not None:
            error_handlers[DecodeError] = on_decode_error
        if on_validation_error is not None:
            error_handlers[ValidationError] = on_validation_error

        return error_handlers



"""
module mqkit.errors

Defines various exception classes used in the MQKit library.
"""

__all__ = [
    "AppError",
    "ConfigurationError",
    "DecodeError",
    "EncodeError",
    "EndpointError",
    "FunctionSignatureError",
    "FunctionTypeError",
    "MarshalError",
    "MessageError",
    "MqkitError",
    "NoForwardTargetError",
    "NoRetry",
    "SerializeError",
    "ShutdownRequested",
    "WorkerTerminatedError",
]

from .apperror import AppError
from .configurationerror import ConfigurationError
from .decodeerror import DecodeError
from .encodeerror import EncodeError
from .endpointerror import EndpointError
from .functionsignatureerror import FunctionSignatureError
from .functiontypeerror import FunctionTypeError
from .marshalerror import MarshalError
from .messageerror import MessageError
from .mqkiterror import MqkitError
from .noforwardtargeterror import NoForwardTargetError
from .noretry import NoRetry
from .serializeerror import SerializeError
from .shutdownrequested import ShutdownRequested
from .workerterminatederror import WorkerTerminatedError



"""
module mqkit.workers.coordinator

Defines the Coordinator abstract base class for automatically managing
endpoints and engines.
"""

from abc import ABCMeta, abstractmethod
from typing import List

from ..endpoints import Endpoint
from ..engines import Engine


class Coordinator(metaclass=ABCMeta):
    """
    class Coordinator

    An abstract base class for coordinators that manage endpoints and engines.
    """

    # pylint: disable=too-few-public-methods

    _endpoints: List[Endpoint]
    _engine: Engine

    def __init__(
        self: "Coordinator", endpoints: List[Endpoint], engine: Engine
    ) -> None:  # pragma: no cover
        self._endpoints = endpoints
        self._engine = engine

    @abstractmethod
    def run(self: "Coordinator") -> None:
        """
        Abstract method to run the coordinator.
        """

        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def stop(self: "Coordinator") -> None:
        """
        Abstract method to stop the coordinator.
        """

        raise NotImplementedError()



"""
module mqkit.connections

Defines connection types and utilities for message queue connections.
Intended for internal use within the mqkit library by the App class
to manage different connection protocols and configurations.
"""

__all__ = ["amqp", "Connection"]

from . import amqp

from .connection import Connection



"""
module mqkit.errors.encodeerror

Defines the EncodeError exception indicating an error during encoding
of a message.
"""

from .marshalerror import MarshalError


class EncodeError(MarshalError):
    """Raised when an error occurs during encoding of a message."""



"""
module mqkit.workers.threaded.threadworker

A module defining a threaded worker for processing messages from a message queue.
"""

import logging
from logging import Logger
from queue import Queue as ProcessQueue, ShutDown
from threading import Event, Thread
from typing import Optional, Union

from ...connections import Connection
from ...endpoints import Endpoint
from ...engines import Engine
from ...errors import NoRetry, ShutdownRequested
from ...logging import root_logger_name
from ...messaging import Forward, QueueMessage
from ...messaging.retry import RetryContext
from .monotoniccounter import MonotonicCounter
from ..worker import Worker


class ThreadWorker(Worker, Thread):
    """
    class ThreadWorker

    A worker that processes messages from a message queue using threading.
    """

    connection: Connection
    error: Optional[Exception] = None

    _counter: MonotonicCounter = MonotonicCounter()

    _endpoint: Endpoint
    _error_queue: Union[ProcessQueue, None]
    _logger: Logger
    _started_event: Event = Event()
    _stopped: bool = False

    def __init__(
        self: "ThreadWorker",
        endpoint: Endpoint,
        engine: Engine,
        error_queue: Union[ProcessQueue, None],
    ) -> None:
        Thread.__init__(
            self, name=f"ThreadWorker-{endpoint.qualname}-{self._counter.next()}"
        )
        Worker.__init__(self)

        self._endpoint = endpoint
        self._engine = engine
        self._error_queue = error_queue
        self._init_logger()

    def _handle_message(self: "ThreadWorker", message: QueueMessage) -> None:
        # pylint: disable=broad-except
        try:
            # process the message using the endpoint's handler. the handler may
            # return a Forward object to indicate that a result should be
            # forwarded to another queue
            forward_result: Optional[Forward] = self._endpoint.handle_message(message)
            self.connection.acknowledge_success(message)

            # forward the result as needed
            if forward_result is not None:
                self._logger.debug(
                    "Forwarding result to %s", forward_result.forward_target
                )
                self.connection.forward_message(
                    forward=forward_result,
                )
        except NoRetry as nr:
            self._logger.warning(
                "Message processing failed with NoRetry%s",
                "" if len(nr.args) == 0 else f": {nr.args[0]}",
            )
            self.connection.acknowledge_failure(message)
        except Exception as exc:
            # log the exception, then allow the retry strategy to determine what action
            # should be taken (e.g. retry, dead-letter, etc.)
            self._logger.exception("Error while processing message: %s", exc)
            self._endpoint.retry_strategy.handle_failure(
                context=RetryContext(
                    connection=self.connection,
                    message=message,
                    exception=exc,
                    received_queue=self._endpoint.queue_name,
                    dead_letter_destination=self._endpoint.dead_letter,
                )
            )

    def _init_logger(self: "ThreadWorker") -> None:
        self._logger = logging.getLogger(
            f"{root_logger_name}.{'.'.join(self.__class__.__module__.split('.')[1:-1])}."
            f"{self.name}"
        )

    def _process_messages(self: "ThreadWorker") -> None:
        while not self._stopped:
            self._handle_message(self._receive_message())

    def _receive_message(self: "ThreadWorker") -> QueueMessage:
        # try getting a message from the queue. this will either block
        # until it succeeds, or raise ShutdownRequested if stop() is called
        message: QueueMessage = self.connection.get_message()
        self._logger.debug(
            "Received message of size %s byte%s",
            f"{len(message.data):,}",
            "" if len(message.data) == 1 else "s",
        )
        return message

    def run(self: "ThreadWorker") -> None:
        # pylint: disable=broad-exception-caught
        try:
            with self._engine.connect(
                queue=self._endpoint.queue_name,
                persistent=self._endpoint.is_persistent,
                auto_delete=self._endpoint.is_auto_delete,
            ) as self.connection:
                self._started_event.set()
                self._process_messages()

        except ShutdownRequested as sr:
            self._logger.warning(
                "Shutdown requested for %s during message processing%s",
                self.name,
                "" if len(sr.args) == 0 else f": {sr.args[0]}",
            )
            self._stopped = True

        except Exception as exc:  # pragma: no cover
            self._logger.exception("Unhandled exception in %s: %s", self.name, exc)
            self.error = exc
            self._stopped = True

        self._started_event.set()
        if self.error is not None and self._error_queue is not None:
            # submit this worker to the error queue if we exited unexpectedly
            try:
                self._error_queue.put(self)
            except ShutDown:
                self._logger.warning(
                    "Failed to submit %s to error queue because it is shut down",
                    self.name,
                )

    def stop(self: "ThreadWorker", message: Optional[str] = None) -> None:
        """
        Stops the worker by unblocking the connection.

        Args:
            message (Optional[str]): An optional message to send when unblocking.

        Returns:
            None

        Raises:
            Nothing
        """

        self._started_event.wait()
        self._stopped = True

        # the while loop in run() may be blocked waiting for a message,
        # so we send a dummy message to unblock it
        #
        # if a message is already processing, the while loop will simply
        # exit after the message is processed
        self.connection.unblock(message=message)

    @property
    def stopped(self: "ThreadWorker") -> bool:
        """
        Indicates whether the worker has been stopped.

        Args:
            None

        Returns:
            bool: True if the worker has been stopped, False otherwise.

        Raises:
            Nothing
        """

        return self._stopped



"""
module mqkit.errors.apperror

Defines the AppError exception indicating an application-level error.
"""

from .mqkiterror import MqkitError


class AppError(MqkitError):
    """Exception indicating an application-level error."""



"""
module mqkit.errors.noretry

Defines the NoRetry exception indicating that the operation should not be retried.
This is intended to be used by user code to signal that a handler for a queue
message should not be retried due to some condition.
"""

from .mqkiterror import MqkitError


class NoRetry(MqkitError):
    """Exception indicating that the operation should not be retried."""



"""
module mqkit.events

Defines the AppEventType enumeration for application event types that
can result in callbacks being invoked.
"""

__all__ = [
    "AppEventType",
]

from .appeventtype import AppEventType



"""
module mqkit.errors.functionsignatureerror

Defines the FunctionSignatureError exception indicating a signature verification
error for a function.
"""

from .endpointerror import EndpointError


class FunctionSignatureError(EndpointError):
    """Raised when there is a signature verification error for a function."""



"""
module mqkit.marshal.serializer

Defines the abstract base class Serializer for serializing and deserializing data
to and from bytestrings using a specified codec.
"""

from abc import ABCMeta, abstractmethod
from typing import Any, Callable, Optional

from .codecs import Codec


class Serializer(metaclass=ABCMeta):
    """
    class Serializer

    Abstract base class for serializers that convert data to and from bytestrings
    using a specific codec.
    """

    _codec: Codec

    def __init__(
        # pylint: disable=unused-argument
        self: "Serializer",
        function: Callable,
        codec: Codec,
    ) -> None:
        self._codec = codec

    @abstractmethod
    def deserialize(self: "Serializer", data: bytes) -> Any:
        """
        Deserialize the given bytestring into data. The returned data should be of
        the appropriate type for the codec associated with this serializer.

        Args:
            data (bytes): The bytestring to deserialize.

        Returns:
            Any: The deserialized data.

        Raises:
            NotImplementedError: If the method is not implemented.
        """

        raise NotImplementedError()

    @abstractmethod
    def serialize(self: "Serializer", data: Any) -> Optional[bytes]:
        """
        Serialize the given data into a bytestring. Data should be of the
        appropriate type for the codec associated with this serializer.

        Args:
            data (Any): The data to serialize.

        Returns:
            Optional[bytes]: The serialized data as a bytestring, or None if
                the data is None.

        Raises:
            NotImplementedError: If the method is not implemented.
        """

        raise NotImplementedError()



"""
module mqkit.marshal.codecs.codec

Defines the abstract base class Codec for encoding and decoding data.
"""

from abc import ABCMeta, abstractmethod
from typing import Any


class Codec(metaclass=ABCMeta):
    """
    class Codec

    An abstract base class for codecs that encode and decode data.
    """

    @property
    @abstractmethod
    def content_type(self: "Codec") -> str:  # pragma: no cover
        """
        Get the content type of the codec.

        Returns:
            str: The content type.

        Raises:
            NotImplementedError: If the method is not implemented.
        """

        raise NotImplementedError()

    @abstractmethod
    def decode(self: "Codec", data: bytes) -> Any:
        """
        Decode the given bytes data into a dictionary.

        Args:
            data (bytes): The bytes data to decode.

        Returns:
            Dict[str, Any]: The decoded data as a dictionary.

        Raises:
            NotImplementedError: If the method is not implemented.
        """

        raise NotImplementedError()

    @abstractmethod
    def encode(self: "Codec", data: Any) -> bytes:
        """
        Encode the given dictionary data into bytes.

        Args:
            data (Dict[str, Any]): The dictionary data to encode.

        Returns:
            bytes: The encoded data as bytes.

        Raises:
            NotImplementedError: If the method is not implemented.
        """

        raise NotImplementedError()



"""
module mqkit.messaging.retry.noretrystrategy

This module implements a retry strategy that never retries a failed message.
This is the default behavior if a retry strategy is not specified for a
message handler.
"""

from typing import override

from ..forward import Forward
from .retrycontext import RetryContext
from .retrystrategy import RetryStrategy


class NoRetryStrategy(RetryStrategy):
    """
    class NoRetryStrategy

    This retry strategy never retries a failed message. This is the default
    behavior if a retry strategy is not specified for a message handler.
    """

    # pylint: disable=too-few-public-methods

    def __init__(self: "NoRetryStrategy") -> None:
        super().__init__()

    @override
    def handle_failure(self: "NoRetryStrategy", context: RetryContext) -> None:
        if context.dead_letter_destination is not None:
            self._logger.info(
                "Message failed, forwarding to dead letter destination %s. Will not retry",
                context.dead_letter_destination,
            )
            context.connection.forward_message(
                Forward(
                    forward_target=context.dead_letter_destination.resource,
                    message=context.message,
                )
            )
        else:
            self._logger.info("Acknowledged failure of message, will not retry")

        # simply acknowledge failure on the existing connection. we never
        # want to retry with this strategy
        context.connection.acknowledge_failure(context.message)



"""
module mqkit.declarations.exchangebinding

Defines the ExchangeBinding class representing a binding between an exchange or queue and
a topic (or routing key in RabbitMQ)
"""

from typing import Union
from pydantic import BaseModel

from ..messaging import Exchange, Queue


class ExchangeBinding(BaseModel):
    """
    class ExchangeBinding

    Represents a binding between an exchange or queue and a topic.
    """

    bound_resource: Union[Exchange, Queue]
    topic: str



"""
module mqkit.declarations

Defines declaration classes for message queue applications.
"""

__all__ = [
    "Declaration",
    "ExchangeBinding",
    "ExchangeDeclaration",
    "QueueDeclaration",
]

from typing import Union

from .exchangebinding import ExchangeBinding
from .exchangedeclaration import ExchangeDeclaration
from .queuedeclaration import QueueDeclaration

Declaration = Union[ExchangeDeclaration, QueueDeclaration]



"""
module mqkit.messaging.forwardtarget

Defines the ForwardTarget type, which represents a target for message forwarding. A ForwardTarget
can be a string (representing a queue name), a Queue object, an Exchange object, or a Destination
object. Apps and @consume marshal this internally to a Destination for forwarding messages.
"""

from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    from ..messaging import Destination, Exchange, Queue

ForwardTarget = Union[str, "Queue", "Exchange", "Destination"]



"""
module mqkit.marshal.codecs

A module providing various codecs for message serialization and deserialization.
"""

__all__ = [
    "Codec",
    "CodecType",
    "JsonCodec",
    "MessagePackCodec",
    "RawCodec",
    "YamlCodec",
]

from .codec import Codec
from .codectype import CodecType
from .jsoncodec import JsonCodec
from .messagepackcodec import MessagePackCodec
from .rawcodec import RawCodec
from .yamlcodec import YamlCodec



"""
module mqkit.messaging.retry

Contains the retry strategy classes for handling failed messages. A retry strategy is responsible
for handling failed messages and determining whether to retry them or not. Subclasses of
RetryStrategy can implement different retry policies. The NoRetryStrategy is a simple
implementation of a retry strategy that never retries a failed message, and is the default
behavior if a retry strategy is not specified for a message handler
"""

__all__ = [
    "ImmediateRetryStrategy",
    "NoRetryStrategy",
    "RetryContext",
    "RetryStrategy",
]

from .immediateretrystrategy import ImmediateRetryStrategy
from .noretrystrategy import NoRetryStrategy
from .retrycontext import RetryContext
from .retrystrategy import RetryStrategy



"""
module mqkit.declarations.exchangedeclaration

Defines the ExchangeDeclaration model representing an exchange declaration.
"""

from typing import List, Union

from pydantic import BaseModel, Field

from ..messaging import Exchange, Queue
from .exchangebinding import ExchangeBinding
from .queuedeclaration import QueueDeclaration


class ExchangeDeclaration(BaseModel):
    """
    class ExchangeDeclaration

    Model representing an exchange declaration.
    """

    exchange: Exchange
    bindings: List[ExchangeBinding] = Field(default_factory=list)

    def bind(
        self: "ExchangeDeclaration",
        resource: Union[Exchange, Queue, "ExchangeDeclaration", QueueDeclaration],
        topic: str,
    ) -> "ExchangeDeclaration":
        """
        Bind a queue to the exchange.

        Args:
            queue (Queue): The queue to bind.

        Returns:
            None
        """

        # unwrap declaration types to an exchange or queue
        if isinstance(resource, ExchangeDeclaration):
            resource = resource.exchange
        if isinstance(resource, QueueDeclaration):
            resource = resource.queue

        self.bindings.append(  # pylint: disable=no-member
            ExchangeBinding(
                bound_resource=resource,
                topic=topic,
            )
        )
        return self

    def bind_exchange(
        self: "ExchangeDeclaration",
        exchange: Union[Exchange, "ExchangeDeclaration", str],
        topic: str,
    ) -> "ExchangeDeclaration":
        """
        Bind an exchange to the exchange.

        Args:
            exchange (Exchange): The exchange to bind.

        Returns:
            None
        """

        if isinstance(exchange, str):
            exchange = Exchange(name=exchange)
        if isinstance(exchange, ExchangeDeclaration):
            exchange = exchange.exchange

        return self.bind(
            resource=exchange,
            topic=topic,
        )

    def bind_queue(
        self: "ExchangeDeclaration",
        queue: Union[Queue, QueueDeclaration, str],
        topic: str,
    ) -> "ExchangeDeclaration":
        """
        Bind a queue to the exchange.

        Args:
            queue (Queue): The queue to bind.

        Returns:
            None
        """

        if isinstance(queue, str):
            queue = Queue(name=queue)
        if isinstance(queue, QueueDeclaration):
            queue = queue.queue

        return self.bind(
            resource=queue,
            topic=topic,
        )



"""
module mqkit.connections.amqp.amqpconsumethread

Defines the AmqpConsumeThread class for consuming AMQP messages in a separate thread.
"""

from threading import Thread
from typing import Optional

from pika.adapters.blocking_connection import BlockingChannel


class AmqpConsumeThread(Thread):
    """
    class AmqpConsumeThread

    A thread that consumes AMQP messages from a BlockingChannel.
    """

    _channel: BlockingChannel
    _exception: Optional[Exception] = None

    def __init__(self, channel: BlockingChannel, **kwargs):
        super().__init__(**kwargs)

        self._channel = channel

    @property
    def error(self: "AmqpConsumeThread") -> Optional[Exception]:  # pragma: no cover
        """
        Gets any exception that occurred during message consumption.

        Args:
            None

        Returns:
            Optional[Exception]: The exception that occurred, or None if no exception
                occurred.

        Raises:
            Nothing
        """

        return self._exception

    def run(self: "AmqpConsumeThread") -> None:  # pragma: no cover
        """
        Starts consuming AMQP messages from the target channel.

        Args:
            None

        Returns:
            None

        Raises:
            Nothing
        """

        # pylint: disable=broad-except
        try:
            self._channel.start_consuming()
        except Exception as ex:
            self._exception = ex
        finally:
            if self._channel.is_open:
                self._channel.close()

    def stop(self: "AmqpConsumeThread") -> None:
        """
        Stops this thread from consuming AMQP messages from the target channel.

        Args:
            None

        Returns:
            None

        Raises:
            Nothing
        """

        self._channel.connection.add_callback_threadsafe(
            self._channel.stop_consuming,
        )



"""
module mqkit.marshal.attributes

Defines the Attributes class representing message attributes in the
message queue system.
"""

from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, Field

from .exceptionhistoryentry import ExceptionHistoryEntry


class Attributes(BaseModel):
    """
    class Attributes

    Represents the attributes of a message in the message queue system.
    """

    headers: Dict[str, str]
    platform: Optional[Dict[str, Any]] = None
    forwarded: bool
    origin_queue: Optional[str] = None
    topic: Union[str, None]
    retry_count: int = 0
    previous_retry_count: int = 0
    is_dead_letter: bool
    exception_history: List[ExceptionHistoryEntry] = Field(default_factory=list)



"""
module mqkit.endpoints.endpoint

Abstract base class for message queue endpoints.
"""

from abc import ABCMeta, abstractmethod
import functools
import inspect
from typing import Any, Callable, Dict, NoReturn, Optional, Type, TypeAlias, Union

from pydantic import ValidationError

from ..errors import FunctionSignatureError, MarshalError
from ..marshal import (
    FullyTypedSerializer,
    ReturnTypeSerializer,
    Serializer,
    TypelessSerializer,
)
from ..marshal.codecs import (
    Codec,
    CodecType,
    JsonCodec,
    MessagePackCodec,
    RawCodec,
    YamlCodec,
)
from ..messaging import Attributes, Destination, Forward, QueueMessage, Response
from ..messaging.retry import RetryStrategy

_codec_type_to_class: Dict[CodecType, Type[Codec]] = {
    CodecType.JSON: JsonCodec,
    CodecType.MESSAGEPACK: MessagePackCodec,
    CodecType.RAW: RawCodec,
    CodecType.YAML: YamlCodec,
}

EndpointCallback: TypeAlias = Callable[
    [Any, Attributes], Optional[Union[Response, bytes]]
]
EndpointExceptionHandler: TypeAlias = Callable[[bytes, Attributes, Exception], None]
EndpointDecodeException: TypeAlias = Union[MarshalError, ValidationError]


class Endpoint(metaclass=ABCMeta):
    """
    class Endpoint

    Abstract base class for message queue endpoints.
    """

    _is_async: bool = False
    _target: Callable[..., Optional[Response]]

    def __init__(self: "Endpoint", target: Callable, codec_type: CodecType) -> None:
        self._is_async = inspect.iscoroutinefunction(target)
        self._target = self._wrap_with_decode(
            target,
            codec_type=codec_type,
        )

    @property
    @abstractmethod
    def dead_letter(self: "Endpoint") -> Optional[Destination]:  # pragma: no cover
        """
        Property that returns the dead letter destination for this endpoint, if any.

        Returns:
            Optional[Destination]: The dead letter destination, or None if not configured.
        """

        raise NotImplementedError(
            f"Endpoint of type {type(self).__name__} does not implement dead_letter()"
        )

    @abstractmethod
    def handle_message(self: "Endpoint", message: QueueMessage) -> Optional[Forward]:
        """
        Handles an incoming message, processes it, and optionally returns a result to
        be forwarded to another message target (queue, topic, etc.)

        Args:
            message (QueueMessage): The incoming message to be processed.

        Returns:
            Optional[Forward]: An optional Forward object indicating where to forward the result.

        Raises:
            NotImplementedError: If the method is not implemented.
        """

        raise NotImplementedError()

    @property
    def is_async(self: "Endpoint") -> bool:
        """
        Property that indicates whether the endpoint's target function is asynchronous.

        Returns:
            bool: True if the target function is asynchronous, False otherwise.
        """

        return self._is_async

    @property
    @abstractmethod
    def is_auto_delete(self: "Endpoint") -> bool:  # pragma: no cover
        """
        Property that indicates whether the endpoint is auto-delete.

        Returns:
            bool: True if the endpoint is auto-delete, False otherwise.
        """

        raise NotImplementedError(
            f"Endpoint of type {type(self).__name__} does not implement is_auto_delete()"
        )

    @property
    @abstractmethod
    def is_persistent(self: "Endpoint") -> bool:  # pragma: no cover
        """
        Property that indicates whether the endpoint is persistent.

        Returns:
            bool: True if the endpoint is persistent, False otherwise.
        """

        raise NotImplementedError(
            f"Endpoint of type {type(self).__name__} does not implement is_persistent()"
        )

    def make_forward_headers(
        self: "Endpoint", response: Response, origin_queue: str
    ) -> Dict[str, str]:
        """
        Creates headers for forwarding a response message.

        Args:
            response (Response): The response message to create headers for.

        Returns:
            Dict[str, str]: A dictionary of headers for the forwarded message.
        """

        return {
            "x-mqkit-forwarded": "true",
            "x-mqkit-origin-queue": origin_queue,
        } | response.headers

    @property
    @abstractmethod
    def qualname(self: "Endpoint") -> str:  # pragma: no cover
        """
        Returns the qualified name of the endpoint.

        Returns:
            str: The qualified name of the endpoint.
        """

        raise NotImplementedError()

    @property
    @abstractmethod
    def queue_name(self: "Endpoint") -> str:
        """
        Property that returns the name of the queue associated with this endpoint.

        Returns:
            str: The name of the queue.
        """

        raise NotImplementedError()  # pragma: no cover

    @property
    @abstractmethod
    def retry_strategy(self: "Endpoint") -> RetryStrategy:
        """
        Property that returns the retry strategy associated with this endpoint.

        Returns:
            RetryStrategy: The retry strategy for the endpoint.
        """

        raise NotImplementedError()  # pragma: no cover

    @property
    def target(self: "Endpoint") -> Callable[..., Optional[Response]]:
        """
        Property that returns the target function associated with this endpoint.

        Returns:
            Callable[..., Any]: The target function.
        """

        return self._target

    def __call__(self, /) -> NoReturn:
        raise TypeError(
            f"{self.target.__name__}() should not be called directly. Use App.run() instead"
        )

    def _make_codec(self: "Endpoint", codec_type: CodecType) -> Codec:
        if codec_type not in _codec_type_to_class:  # pragma: no cover
            raise NotImplementedError(f"Unsupported codec type: {codec_type}")

        return _codec_type_to_class[codec_type]()

    def _make_serializer(
        self: "Endpoint", func: EndpointCallback, codec_type: CodecType
    ) -> Serializer:
        codec: Codec = self._make_codec(codec_type)

        signature: inspect.Signature = inspect.signature(func)
        if len(signature.parameters) != 2:
            raise FunctionSignatureError(
                f"Function {func.__name__}() must accept exactly two parameters"
            )

        # check for no type annotations. if this is the case, infer no return value
        if func.__annotations__ == {}:
            return self._make_serializer_typeless(function=func, codec=codec)

        # check for a return type annotation
        if "return" in func.__annotations__ and len(func.__annotations__) == 1:
            return self._make_serializer_return_type(function=func, codec=codec)

        # just try constructing a fully-typed serializer. it will raise an appropriate
        # FunctionSignatureError if it can't make sense of the signature
        return self._make_serializer_fully_typed(function=func, codec=codec)

    def _make_serializer_fully_typed(
        self: "Endpoint",
        function: Callable,
        codec: Codec,
    ) -> Serializer:
        return FullyTypedSerializer(function=function, codec=codec)

    def _make_serializer_return_type(
        self: "Endpoint",
        function: Callable,
        codec: Codec,
    ) -> Serializer:
        return ReturnTypeSerializer(function=function, codec=codec)

    def _make_serializer_typeless(
        self: "Endpoint",
        function: Callable,
        codec: Codec,
    ) -> Serializer:
        return TypelessSerializer(function=function, codec=codec)

    def _wrap_with_decode(
        self: "Endpoint",
        func: EndpointCallback,
        codec_type: CodecType,
    ) -> Callable[..., Optional[Response]]:
        serializer: Serializer = self._make_serializer(
            func=func,
            codec_type=codec_type,
        )

        @functools.wraps(func)
        def _deserialize_wrapper(
            *,
            message: bytes,
            attributes: Attributes,
        ) -> Optional[Response]:
            # call the function and get the result. it could be a Response object, some
            # data, or None. we want to marshal this into Response | None
            result: Optional[Union[Response, bytes]] = func(
                serializer.deserialize(message),
                attributes,
            )
            if not isinstance(result, Response):
                result = Response(content=result)

            # serialize the content into byte data and return the Response. if the serializer
            # returns None, Response.data will raise an error
            serialized_data: Optional[bytes] = serializer.serialize(result.content)
            if serialized_data is None:
                return None
            result.data = serialized_data
            return result

        return _deserialize_wrapper



"""
module mqkit.endpoints

Defines the Endpoint and QueueEndpoint classes for message queue endpoints.
"""

__all__ = ["config", "Endpoint", "EndpointFactory", "QueueEndpoint"]


from . import config

from .endpoint import Endpoint
from .endpointfactory import EndpointFactory
from .queueendpoint import QueueEndpoint



"""
module mqkit.marshal.fullytypedserializer

Defines the FullyTypedSerializer for serializing function inputs and outputs with
full type information.
"""

import inspect
from typing import Any, Callable, Dict, Optional, Type, override

from pydantic import BaseModel

from .codecs.codec import Codec
from ..errors import FunctionSignatureError
from ..messaging import Attributes
from .returntypeserializer import ReturnTypeSerializer


class FullyTypedSerializer(ReturnTypeSerializer):
    """
    class FullyTypedSerializer

    Serializer that enforces full type annotations on function parameters and return type.
    """

    _message_type: Optional[Type[BaseModel]] = None

    def __init__(
        self: "FullyTypedSerializer",
        function: Callable,
        codec: Codec,
    ) -> None:
        # call the ReturnTypeSerializer constructor to check the return type
        super().__init__(
            function=function,
            codec=codec,
        )

        self._check_function_signature(function)

    @override
    def deserialize(self: "FullyTypedSerializer", data: bytes) -> Any:
        result: Any = self._codec.decode(data)

        if self._message_type is not None:
            if isinstance(result, self._message_type):
                return result
            if isinstance(result, Dict):
                return self._message_type(**result)

            raise TypeError(
                f"Deserialized data is of type {type(result).__name__}, "
                f"expected {self._message_type.__name__} or Dict"
            )

        if isinstance(result, Dict):
            return result

        raise TypeError(
            f"Codec returned value of type {type(result).__name__}, expected Dict or model"
        )

    @override
    def serialize(self: "FullyTypedSerializer", data: Any) -> Optional[bytes]:
        # since we inherit from ReturnTypeSerializer, we can just use its serialize logic
        return super().serialize(data)

    def _check_function_signature(
        self: "FullyTypedSerializer",
        function: Callable,
    ) -> None:
        signature: inspect.Signature = inspect.signature(function)

        self._check_type_annotations(function, signature)

    def _check_type_annotations(
        self: "FullyTypedSerializer",
        function: Callable,
        signature: inspect.Signature,
    ) -> None:
        if len(signature.parameters) != 2:  # pragma: no cover
            raise FunctionSignatureError(
                f"Function {function.__name__}() must accept exactly two parameters"
            )

        for ann_name in (
            "message",
            "attributes",
            "return",
        ):
            if ann_name not in function.__annotations__:
                raise FunctionSignatureError(
                    f"Function {function.__name__}() must have type annotation for '{ann_name}'"
                )

        # check the message parameter type and populate the message model type
        self._populate_message_model_type(
            function,
            signature.parameters["message"].annotation,
        )

        if signature.parameters["attributes"].annotation is not Attributes:
            raise FunctionSignatureError(
                f"Function {function.__name__}() 'attributes' parameter type "
                f"must be mqkit.messaging.Attributes"
            )

    def _populate_message_model_type(
        self: "FullyTypedSerializer",
        function: Callable,
        annotation: Type,
    ) -> None:
        if issubclass(annotation, Dict):
            return

        if not issubclass(annotation, BaseModel):
            raise FunctionSignatureError(
                f"Function {function.__name__}() 'message' parameter type "
                f"must be a subclass of pydantic.BaseModel"
            )

        self._message_type = annotation
